import torch
import torch.nn as nn
import torch.nn.functional as F
import ltr.models.backbone as backbones
import ltr.models.segm as segmmodels
import ltr.models.bbox_reg as bbox_reg_models
from ltr import model_constructor
from ltr import load_network

from ltr.admin.loading import load_existing_network



class BboxSegmNet(nn.Module):
    def __init__(self, feature_extractor, segm_predictor, bbox_predictor, segm_layers, extractor_grad=True):
        super(BboxSegmNet, self).__init__()

        self.feature_extractor = feature_extractor
        self.segm_predictor = segm_predictor
        self.bbox_predictor = bbox_predictor
        self.segm_layers = segm_layers

        # do not train backbone
        if not extractor_grad:
            for p in self.feature_extractor.parameters():
                p.requires_grad_(False)

        # do not train segmentation network
        for p in self.segm_predictor.parameters():
            p.requires_grad_(False)

    def forward(self, train_imgs, test_imgs, train_masks, test_dist, train_dist):

        # extract backbone features
        train_feat = self.extract_backbone_features(train_imgs)
        test_feat = self.extract_backbone_features(test_imgs)

        # prepare inputs for segmentation
        train_feat_segm = [feat for feat in train_feat.values()]
        test_feat_segm = [feat for feat in test_feat.values()]
        train_masks = [train_masks]
        test_dist = [test_dist]
        train_dist = [train_dist]

        # Obtain segmentation prediction
        # segm_pred = self.segm_predictor(test_feat_segm, train_feat_segm, train_masks, test_dist)

        # predict segmentation mask from bounding box on training image
        segm_pred_train, _ = self.segm_predictor(train_feat_segm, train_feat_segm, train_masks, train_dist)
        mask_pred_train = [torch.unsqueeze(F.softmax(segm_pred_train, dim=1)[:, 0, :, :], dim=1)]

        # use segmentation prediction on training image and predict segmentation mask on test image
        segm_pred, _ = self.segm_predictor(test_feat_segm, train_feat_segm, mask_pred_train, test_dist)

        mask = torch.unsqueeze(F.softmax(segm_pred, dim=1)[:, 0, :, :], dim=1)

        reg, cls = self.bbox_predictor(train_feat_segm, test_feat_segm, mask)

        return reg, cls, mask

    def extract_backbone_features(self, im, layers=None):
        if layers is None:
            layers = self.segm_layers
        return self.feature_extractor(im, layers)

    def extract_features(self, im, layers):
        return self.feature_extractor(im, layers)


@model_constructor
def bbox_reg_resnet50(segm_input_dim=(256,256), segm_inter_dim=(256,256), backbone_pretrained=True,
                      topk_pos=3, topk_neg=3, mixer_channels=2):
    # backbone
    # backbone_net = backbones.resnet50(pretrained=backbone_pretrained)
    backbone_net = backbones.resnet50_octconv()
    if backbone_pretrained is not None:
        load_existing_network(backbone_net, backbone_pretrained)

    segm_net_path = '/home/benjamin/Desktop/D3S/SegmNet_ep0040.pth.tar'
    segm_net, _ = load_network(segm_net_path, backbone_pretrained=backbone_pretrained,
                               constructor_module='ltr.models.segm.segm',
                               constructor_fun_name='segm_resnet50')

    if segm_net is None:
        raise RuntimeError('Segmentation network is not given.')
    segm_predictor = segm_net.segm_predictor

    # parameters: dimensions for bounding box prediction
    in_channels = 1024
    project_channels = 256
    dxcorr_channels = 256
    mask_channels = 16
    dxcorr_adj_channels = 32
    mixer_channels = 32

    bbox_predictor = bbox_reg_models.BboxRegNet(in_channels=in_channels,
                                                project_channels=project_channels,
                                                dxcorr_channels=dxcorr_channels,
                                                mask_channels=mask_channels,
                                                dxcorr_adj_channels=dxcorr_adj_channels,
                                                mixer_channels=mixer_channels)

    net = BboxSegmNet(feature_extractor=backbone_net, segm_predictor=segm_predictor, bbox_predictor=bbox_predictor,
                      segm_layers=['conv1', 'layer1', 'layer2', 'layer3'], extractor_grad=False)

    return net
