import torch.nn as nn
import torch
import torch.nn.functional as F

from ltr.models.backbone import Bottleneck
from ltr.ops.octconv.octconv import Conv_BN


def conv(in_planes, out_planes, kernel_size=3, stride=1, padding=1, dilation=1):
    return nn.Sequential(
            nn.Conv2d(in_planes, out_planes, kernel_size=kernel_size, stride=stride,
                      padding=padding, dilation=dilation, bias=True),
            nn.BatchNorm2d(out_planes),
            nn.ReLU(inplace=True))

def conv_no_relu(in_planes, out_planes, kernel_size=3, stride=1, padding=1, dilation=1):
    return nn.Sequential(
            nn.Conv2d(in_planes, out_planes, kernel_size=kernel_size, stride=stride,
                      padding=padding, dilation=dilation, bias=True),
            nn.BatchNorm2d(out_planes))

def depthwise_corr(x, kernel):
    """ depthwise cross correlation """
    batch = kernel.size(0)
    channel = kernel.size(1)
    x = x.view(1, batch*channel, x.size(2), x.size(3))
    kernel = kernel.view(batch*channel, 1, kernel.size(2), kernel.size(3))
    out = F.conv2d(x, kernel, groups=batch*channel)
    out = out[:, :, :-1, :-1]
    out = out.view(batch, channel, out.size(2), out.size(3))
    return out

class DXCorrModule(nn.Module):
    def __init__(self, in_channels, hidden, mask_channels, dxcorr_adj_channels, mixer_channels, out_channels):
        super(DXCorrModule, self).__init__()

        self.kernel_conv = nn.Sequential(
            nn.Conv2d(in_channels, hidden, kernel_size=3, bias=False, padding=1),
            nn.BatchNorm2d(hidden),
            nn.ReLU(inplace=True),
        )
        self.search_conv = nn.Sequential(
            nn.Conv2d(in_channels, hidden, kernel_size=3, bias=False, padding=1),
            nn.BatchNorm2d(hidden),
            nn.ReLU(inplace=True),
        )
        self.adjust_dxcorr = nn.Sequential(
            nn.Conv2d(hidden, dxcorr_adj_channels, kernel_size=1, bias=False),
            nn.BatchNorm2d(dxcorr_adj_channels),
        )
        self.corr_mask_mixer = nn.Sequential(
            nn.Conv2d(dxcorr_adj_channels + mask_channels, mixer_channels, kernel_size=3, bias=False, padding=1),
            nn.BatchNorm2d(mixer_channels),
            nn.ReLU(inplace=True),
        )
        self.post_conv = nn.Sequential(
            nn.Conv2d(mixer_channels, mixer_channels, kernel_size=1, bias=False),  # mixer_channels
            nn.BatchNorm2d(mixer_channels),  # mixer_channels
            nn.ReLU(inplace=True),
            nn.Conv2d(mixer_channels, out_channels, kernel_size=1)
        )

    def forward(self, features_search, features_kernel, features_mask_search):
        kernel = self.kernel_conv(features_kernel)
        search = self.search_conv(features_search)

        # take central part of the kernel
        kernel = kernel[:, :, 8:16, 8:16]  # if features are from conv-3
        # kernel = kernel[:, :, 18:30, 18:30]  # if features are from conv-2

        dxcorr = depthwise_corr(search, kernel)
        dxcorr = self.adjust_dxcorr(dxcorr)

        d_ = (features_mask_search.shape[-1] - dxcorr.shape[-1]) // 2
        features_mask_search = features_mask_search[:, :, d_:-d_, d_:-d_]

        dxcorr_mask = torch.cat((dxcorr, features_mask_search), dim=1)

        dxcorr_mask = self.corr_mask_mixer(dxcorr_mask)

        out = self.post_conv(dxcorr_mask)

        return out

class MaskResampleModule(nn.Module):
    def __init__(self, out_channels):
        super(MaskResampleModule, self).__init__()

        self.reduce_1 = nn.Sequential(
            nn.Conv2d(1, 4, kernel_size=3, bias=False, stride=2, padding=1),
            nn.BatchNorm2d(4),
            nn.ReLU(inplace=True),
        )
        self.reduce_2 = nn.Sequential(
            nn.Conv2d(4, 4, kernel_size=3, bias=False, stride=2, padding=1),
            nn.BatchNorm2d(4),  # 4
            nn.ReLU(inplace=True),
        )
        self.reduce_3 = nn.Sequential(
            nn.Conv2d(4, 8, kernel_size=3, bias=False, stride=2, padding=1),
            nn.BatchNorm2d(8),  # 8
            nn.ReLU(inplace=True),
        )
        self.reduce_4 = nn.Sequential(
            nn.Conv2d(8, out_channels, kernel_size=3, bias=False, stride=2, padding=1),
            nn.BatchNorm2d(out_channels),
        )

    def forward(self, mask_search):
        # mask_search: 384 x 384
        out = self.reduce_1(mask_search)  # 192
        out = self.reduce_2(out)  # 96
        out = self.reduce_3(out)  # 48
        out = self.reduce_4(out)  # 24

        return out

class BboxRegNet(nn.Module):
    def __init__(self, in_channels, project_channels, dxcorr_channels, mask_channels, dxcorr_adj_channels, mixer_channels):
        super().__init__()
        self.inplanes = 128
        self.groups = 1
        self.base_width = 64

        self.project = self._make_layer(Bottleneck, in_channels, project_channels // 4, 2, ### "1x1 conv, 256" <- reduced to 256?
                                        stride=1, norm_layer=nn.BatchNorm2d,
                                        alpha_out=0, output=True)

        self.mask_conv = MaskResampleModule(mask_channels) ### "Mask adjustment module"?

        self.dx_corr_reg = DXCorrModule(project_channels, dxcorr_channels, mask_channels, dxcorr_adj_channels, mixer_channels, 4)
        self.dx_corr_cls = DXCorrModule(project_channels, dxcorr_channels, mask_channels, dxcorr_adj_channels, mixer_channels, 2)

        # Init weights
        for m in self.modules():
            if isinstance(m, nn.Conv2d) or isinstance(m, nn.ConvTranspose2d) or isinstance(m, nn.Linear):
                nn.init.kaiming_normal_(m.weight.data, mode='fan_in')
                if m.bias is not None:
                    m.bias.data.zero_()

        self.exp_param = nn.Parameter(torch.ones(1))

    def _make_layer(self, block, inplanes, planes, blocks, stride=1, alpha_in=0.5, alpha_out=0.5, norm_layer=None,
                    output=False):
        if norm_layer is None:
            norm_layer = nn.BatchNorm2d
        downsample = None
        if stride != 1 or inplanes != planes * block.expansion:
            downsample = nn.Sequential(
                Conv_BN(inplanes, planes * block.expansion, kernel_size=1, stride=stride, alpha_in=alpha_in,
                        alpha_out=alpha_out)
            )

        layers = []
        layers.append(block(inplanes, planes, stride, downsample, self.groups, ### add 1 Bottleneck layer? why?
                            self.base_width, alpha_in, alpha_out, norm_layer, output))
        inplanes = planes * block.expansion
        for _ in range(1, blocks): ### range 1,2 - add 1 Bottleneck layer?
            layers.append(block(inplanes, planes, groups=self.groups,
                                base_width=self.base_width, norm_layer=norm_layer,
                                alpha_in=0 if output else 0.5, alpha_out=0 if output else 0.5, output=output))

        return nn.Sequential(*layers)

    def forward(self, feat_template, feat_search, mask_search):
        """
        Bounding box regression forward pass
        :param feat_template:
        :param feat_search:
        :param mask_search:
        :return:
        """

        # lower dimensionality of the features and specialize them for bbox regression
        ft = self.project(feat_template[3])[0]
        fs = self.project(feat_search[3])[0]

        # create mask features from the mask
        fm = self.mask_conv(mask_search)

        # depth-wise correlation to get final predictions
        reg = self.dx_corr_reg(fs, ft, fm)  # B x 4 x H x W
        cls = self.dx_corr_cls(fs, ft, fm)  # B x 2 x H x W

        reg = torch.exp(reg * self.exp_param)

        return reg, cls
