import torch.nn as nn
import torch
import torch.nn.functional as F
import numpy as np

from ltr.models.backbone import Bottleneck
from ltr.ops.octconv.octconv import Conv_BN_ACT, Conv_BN


def conv(in_planes, out_planes, kernel_size=3, stride=1, padding=1, dilation=1):
    return nn.Sequential(
            nn.Conv2d(in_planes, out_planes, kernel_size=kernel_size, stride=stride,
                      padding=padding, dilation=dilation, bias=True),
            nn.BatchNorm2d(out_planes),
            nn.ReLU(inplace=True))

def conv_no_relu(in_planes, out_planes, kernel_size=3, stride=1, padding=1, dilation=1):
    return nn.Sequential(
            nn.Conv2d(in_planes, out_planes, kernel_size=kernel_size, stride=stride,
                      padding=padding, dilation=dilation, bias=True),
            nn.BatchNorm2d(out_planes))


def valid_roi(roi: torch.Tensor, image_size: torch.Tensor):
    valid = all(0 <= roi[:, 1]) and all(0 <= roi[:, 2]) and all(roi[:, 3] <= image_size[0]-1) and \
            all(roi[:, 4] <= image_size[1]-1)
    return valid


def normalize_vis_img(x):
    x = x - np.min(x)
    x = x / np.max(x)
    return (x * 255).astype(np.uint8)

def upsample_twice(input_):
    return F.interpolate(input_, scale_factor=2, mode='bilinear', align_corners=True)

def upsample_ca(x, backbone_f):
    x_pool = F.adaptive_avg_pool2d(x, (1, 1))
    features_pool = F.adaptive_avg_pool2d(backbone_f, (1, 1))
    return upsample_twice(x * x_pool + backbone_f * features_pool)

class SegmNet(nn.Module):
    """ Network module for IoU prediction. Refer to the paper for an illustration of the architecture."""
    def __init__(self, segm_input_dim=(128,256), segm_inter_dim=(256,256), segm_dim=(64, 64), mixer_channels=2, topk_pos=3, topk_neg=3):
        super().__init__()

        self.segment1 = conv_no_relu(segm_dim[0], segm_dim[1])

        self.pos_conv_0 = conv(24 * 24, 16, kernel_size=1, padding=0)
        self.neg_conv_0 = conv(24 * 24, 16, kernel_size=1, padding=0)
        self.pos_conv_1 = nn.Conv2d(16, 1, kernel_size=1, stride=1, padding=0, dilation=1, bias=True)
        self.neg_conv_1 = nn.Conv2d(16, 1, kernel_size=1, stride=1, padding=0, dilation=1, bias=True)

        self.mixer = conv(mixer_channels, segm_inter_dim[3])

        self.s3 = conv(segm_inter_dim[3], segm_inter_dim[3])
        self.s2 = conv(segm_inter_dim[2], segm_inter_dim[2])
        self.s1 = conv(segm_inter_dim[1], segm_inter_dim[1])
        self.s0 = conv(segm_inter_dim[0], segm_inter_dim[0])

        self.f0 = conv(segm_input_dim[0], segm_inter_dim[0])

        self.post3 = conv(segm_inter_dim[3], segm_inter_dim[2])
        self.post2 = conv(segm_inter_dim[2], segm_inter_dim[1])
        self.post1 = conv(segm_inter_dim[1], segm_inter_dim[0])
        self.post0 = conv_no_relu(segm_inter_dim[0], 2)

        # for octave convolution
        norm_layer = nn.BatchNorm2d

        self.inplanes = 128
        self.groups = 1
        self.base_width = 64

        self.segment0 = self._make_layer(Bottleneck, segm_input_dim[3], segm_dim[0] // 4, 2,
                                         stride=1, norm_layer=norm_layer,
                                         alpha_out=0, output=True)

        self.f3 = self._make_layer(Bottleneck, segm_input_dim[3], segm_inter_dim[3] // 4, 2,
                                   stride=1, norm_layer=norm_layer,
                                   alpha_out=0, output=True)
        self.f2 = self._make_layer(Bottleneck, segm_input_dim[2], segm_inter_dim[2] // 4, 2,
                                   stride=1, norm_layer=norm_layer,
                                   alpha_out=0, output=True)
        self.f1 = self._make_layer(Bottleneck, segm_input_dim[1], segm_inter_dim[1] // 4, 2,
                                   stride=1, norm_layer=norm_layer,
                                   alpha_out=0, output=True)

        # Init weights
        for m in self.modules():
            if isinstance(m, nn.Conv2d) or isinstance(m, nn.ConvTranspose2d) or isinstance(m, nn.Linear):
                nn.init.kaiming_normal_(m.weight.data, mode='fan_in')
                if m.bias is not None:
                    m.bias.data.zero_()

    def _make_layer(self, block, inplanes, planes, blocks, stride=1, alpha_in=0.5, alpha_out=0.5, norm_layer=None,
                    output=False):
        if norm_layer is None:
            norm_layer = nn.BatchNorm2d
        downsample = None
        if stride != 1 or inplanes != planes * block.expansion:
            downsample = nn.Sequential(
                Conv_BN(inplanes, planes * block.expansion, kernel_size=1, stride=stride, alpha_in=alpha_in,
                        alpha_out=alpha_out)
            )

        layers = []
        layers.append(block(inplanes, planes, stride, downsample, self.groups,
                            self.base_width, alpha_in, alpha_out, norm_layer, output))
        inplanes = planes * block.expansion
        for _ in range(1, blocks):
            layers.append(block(inplanes, planes, groups=self.groups,
                                base_width=self.base_width, norm_layer=norm_layer,
                                alpha_in=0 if output else 0.5, alpha_out=0 if output else 0.5, output=output))

        return nn.Sequential(*layers)

    def forward(self, feat_test, feat_train, mask_train, test_dist=None):
        f_test = self.segment1(self.segment0(feat_test[3])[0])
        f_train = self.segment1(self.segment0(feat_train[3])[0])

        # reshape mask to the feature size
        mask_pos = F.interpolate(mask_train[0], size=(f_train.shape[-2], f_train.shape[-1]))
        mask_neg = 1 - mask_pos

        pred_pos, pred_neg = self.similarity_segmentation(f_test, f_train, mask_pos, mask_neg) ### GIM?

        pred_ = torch.cat((torch.unsqueeze(pred_pos, -1), torch.unsqueeze(pred_neg, -1)), dim=-1)
        pred_sm = F.softmax(pred_, dim=-1)

        if test_dist is not None:
            # distance map is give - resize for mixer
            dist = F.interpolate(test_dist[0], size=(f_train.shape[-2], f_train.shape[-1]))
            # concatenate inputs for mixer
            # softmaxed segmentation, positive segmentation and distance map
            segm_layers = torch.cat((torch.unsqueeze(pred_sm[:, :, :, 0], dim=1),
                                     torch.unsqueeze(pred_pos, dim=1),
                                     dist), dim=1)
        else:
            segm_layers = torch.cat((torch.unsqueeze(pred_sm[:, :, :, 0], dim=1), torch.unsqueeze(pred_pos, dim=1)), dim=1)

        out = self.mixer(segm_layers)

        out = self.post3(upsample_ca(self.s3(out), self.f3(feat_test[3])[0]))
        out = self.post2(upsample_ca(self.s2(out), self.f2(feat_test[2])[0]))
        out = self.post1(upsample_ca(self.s1(out), self.f1(feat_test[1])[0]))
        out = self.post0(upsample_ca(self.s0(out), self.f0(feat_test[0])))

        return out, pred_.permute(0, 3, 1, 2)


    def similarity_segmentation(self, f_test, f_train, mask_pos, mask_neg):
        # first normalize train and test features to have L2 norm 1
        # cosine similarity and reshape last two dimensions into one
        sim = torch.einsum('ijkl,ijmn->iklmn',
                           F.normalize(f_test, p=2, dim=1),
                           F.normalize(f_train, p=2, dim=1))

        sim_resh = sim.view(sim.shape[0], sim.shape[1], sim.shape[2], sim.shape[3] * sim.shape[4])

        # reshape masks into vectors for broadcasting [B x 1 x 1 x w * h]
        # re-weight samples (take out positive ang negative samples)
        sim_pos = sim_resh * mask_pos.view(mask_pos.shape[0], 1, 1, -1)
        sim_neg = sim_resh * mask_neg.view(mask_neg.shape[0], 1, 1, -1)

        # sort positive and negative similarity tensors and make 1x1 convolution instead topk mean
        pos_map = torch.squeeze(self.pos_conv_1(self.pos_conv_0(torch.sort(sim_pos, dim=-1)[0].permute(0, 3, 1, 2))), dim=1)
        neg_map = torch.squeeze(self.neg_conv_1(self.neg_conv_0(torch.sort(sim_neg, dim=-1)[0].permute(0, 3, 1, 2))), dim=1)

        return pos_map, neg_map

