import functools
import torch


class TensorTuple(tuple):
    """Container mainly used for tuples of torch tensors. Extends tuples with pytorch functionality."""

    def __new__(cls, *args, **kwargs):
        return tuple.__new__(cls, args)

    def __getitem__(self, item):
        if isinstance(item, int):
            return super(TensorTuple, self).__getitem__(item)
        elif isinstance(item, (tuple, list)):
            return TensorTuple(*[super(TensorTuple, self).__getitem__(i) for i in item])
        else:
            return TensorTuple(super(TensorTuple, self).__getitem__(item))

    def __add__(self, other):
        if TensorTuple._iterable(other):
            return TensorTuple(*[e1 + e2 for e1, e2 in zip(self, other)])
        return TensorTuple(*[e + other for e in self])

    def __radd__(self, other):
        if TensorTuple._iterable(other):
            return TensorTuple(*[e2 + e1 for e1, e2 in zip(self, other)])
        return TensorTuple(*[other + e for e in self])

    def __iadd__(self, other):
        if TensorTuple._iterable(other):
            for i, e2 in enumerate(other):
                self[i] += e2
        else:
            for i in range(len(self)):
                self[i] += other
        return self

    def __sub__(self, other):
        if TensorTuple._iterable(other):
            return TensorTuple(*[e1 - e2 for e1, e2 in zip(self, other)])
        return TensorTuple(*[e - other for e in self])

    def __rsub__(self, other):
        if TensorTuple._iterable(other):
            return TensorTuple(*[e2 - e1 for e1, e2 in zip(self, other)])
        return TensorTuple(*[other - e for e in self])

    def __isub__(self, other):
        if TensorTuple._iterable(other):
            for i, e2 in enumerate(other):
                self[i] -= e2
        else:
            for i in range(len(self)):
                self[i] -= other
        return self

    def __mul__(self, other):
        if TensorTuple._iterable(other):
            return TensorTuple(*[e1 * e2 for e1, e2 in zip(self, other)])
        return TensorTuple(*[e * other for e in self])

    def __rmul__(self, other):
        if TensorTuple._iterable(other):
            return TensorTuple(*[e2 * e1 for e1, e2 in zip(self, other)])
        return TensorTuple(*[other * e for e in self])

    def __imul__(self, other):
        if TensorTuple._iterable(other):
            for i, e2 in enumerate(other):
                self[i] *= e2
        else:
            for i in range(len(self)):
                self[i] *= other
        return self

    def __truediv__(self, other):
        if TensorTuple._iterable(other):
            return TensorTuple(*[e1 / e2 for e1, e2 in zip(self, other)])
        return TensorTuple(*[e / other for e in self])

    def __rtruediv__(self, other):
        if TensorTuple._iterable(other):
            return TensorTuple(*[e2 / e1 for e1, e2 in zip(self, other)])
        return TensorTuple(*[other / e for e in self])

    def __itruediv__(self, other):
        if TensorTuple._iterable(other):
            for i, e2 in enumerate(other):
                self[i] /= e2
        else:
            for i in range(len(self)):
                self[i] /= other
        return self

    def __matmul__(self, other):
        if TensorTuple._iterable(other):
            return TensorTuple(*[e1 @ e2 for e1, e2 in zip(self, other)])
        return TensorTuple(*[e @ other for e in self])

    def __rmatmul__(self, other):
        if TensorTuple._iterable(other):
            return TensorTuple(*[e2 @ e1 for e1, e2 in zip(self, other)])
        return TensorTuple(*[other @ e for e in self])

    def __imatmul__(self, other):
        if TensorTuple._iterable(other):
            for i, e2 in enumerate(other):
                self[i] @= e2
        else:
            for i in range(len(self)):
                self[i] @= other
        return self

    def __mod__(self, other):
        if TensorTuple._iterable(other):
            return TensorTuple(*[e1 % e2 for e1, e2 in zip(self, other)])
        return TensorTuple(*[e % other for e in self])

    def __rmod__(self, other):
        if TensorTuple._iterable(other):
            return TensorTuple(*[e2 % e1 for e1, e2 in zip(self, other)])
        return TensorTuple(*[other % e for e in self])

    def __pos__(self):
        return TensorTuple(*[+e for e in self])

    def __neg__(self):
        return TensorTuple(*[-e for e in self])

    def __le__(self, other):
        if TensorTuple._iterable(other):
            return TensorTuple(*[e1 <= e2 for e1, e2 in zip(self, other)])
        return TensorTuple(*[e <= other for e in self])

    def __ge__(self, other):
        if TensorTuple._iterable(other):
            return TensorTuple(*[e1 >= e2 for e1, e2 in zip(self, other)])
        return TensorTuple(*[e >= other for e in self])

    def concat(self, other):
        return TensorTuple(super(TensorTuple, self).__add__(other))

    def copy(self):
        return TensorTuple(*[t for t in self])

    def clone(self):
        return TensorTuple(*[t.clone() for t in self])

    def unroll(self):
        if not any(isinstance(t, TensorTuple) for t in self):
            return self

        new_list = TensorTuple()
        for t in self:
            if isinstance(t, TensorTuple):
                new_list.extend(t.unroll())
            else:
                new_list.append(t)
        return new_list

    def attribute(self, attr: str, *args):
        return TensorTuple(*[getattr(e, attr, *args) for e in self])

    def apply(self, fn):
        return TensorTuple(*[fn(e) for e in self])

    def __getattr__(self, name):
        if not hasattr(torch.Tensor, name):
            raise AttributeError('\'TensorTuple\' object has not attribute \'{}\''.format(name))

        def apply_attr(*args, **kwargs):
            return TensorTuple(*[getattr(e, name)(*args, **kwargs) for e in self])

        return apply_attr

    @staticmethod
    def _iterable(a):
        return isinstance(a, (TensorTuple, list))



def tensor_operation(op):
    def islist(a):
        return isinstance(a, TensorTuple)

    @functools.wraps(op)
    def oplist(*args, **kwargs):
        if len(args) == 0:
            raise ValueError('Must be at least one argument without keyword (i.e. operand).')

        if len(args) == 1:
            if islist(args[0]):
                return TensorTuple(*[op(a, **kwargs) for a in args[0]])
        else:
            # Multiple operands, assume max two
            if islist(args[0]) and islist(args[1]):
                return TensorTuple(*[op(a, b, *args[2:], **kwargs) for a, b in zip(*args[:2])])
            if islist(args[0]):
                return TensorTuple(*[op(a, *args[1:], **kwargs) for a in args[0]])
            if islist(args[1]):
                return TensorTuple(*[op(args[0], b, *args[2:], **kwargs) for b in args[1]])

        # None of the operands are lists
        return op(*args, **kwargs)

    return oplist