import os
import sys
import argparse
import time

import cv2
import numpy as np

env_path = os.path.join(os.path.dirname(__file__), '..')
if env_path not in sys.path:
    sys.path.append(env_path)

from vot.dataset import VOTDataset
from vot.region.shapes import Rectangle
from vot.region.raster import calculate_overlap

import copy
from pytracking.tracker.segm import Segm
from pytracking.parameter.segm import default_params as params


def draw_text(img, text, text_pos):
    font = cv2.FONT_HERSHEY_PLAIN
    text_sz = cv2.getTextSize(text, font, 1, 1)
    tl_ = (text_pos[0] - 5, text_pos[1] + 5)
    br_ = (text_pos[0] - 5 + text_sz[0][0] + 10, text_pos[1] - 5 - text_sz[0][1])
    cv2.rectangle(img, tl_, br_, (0, 0, 0), cv2.FILLED)
    cv2.putText(img, text, text_pos, font, 1, (255, 255, 255), 1, cv2.LINE_AA, False)

def draw_progress_bar(img, frame_index, sequence_length):
    cv2.line(img, (0, img.shape[0]), (img.shape[1], img.shape[0]), (0, 0, 0), 4)
    end_point = int(round((frame_index / sequence_length) * img.shape[1]))
    cv2.line(img, (0, img.shape[0]), (end_point, img.shape[0]), (255, 255, 255), 4)

def overlay_mask(img, mask, a=0.4, mask_color=(0, 255, 0), contour_line_width=2):
    if a >= 0:
        img_b = img[:, :, 0]
        img_g = img[:, :, 1]
        img_r = img[:, :, 2]

        img_r[mask > 0] = np.round(img_r[mask > 0] * (1 - a) + a * mask_color[0])
        img_g[mask > 0] = np.round(img_g[mask > 0] * (1 - a) + a * mask_color[1])
        img_b[mask > 0] = np.round(img_b[mask > 0] * (1 - a) + a * mask_color[2])

    M = (mask > 0).astype(np.uint8)
    if cv2.__version__[-5] == '4':
        contours, _ = cv2.findContours(M, cv2.RETR_EXTERNAL, cv2.CHAIN_APPROX_NONE)
    else:
        _, contours, _ = cv2.findContours(M, cv2.RETR_EXTERNAL, cv2.CHAIN_APPROX_NONE)

    cv2.drawContours(img, contours, -1, mask_color[::-1], contour_line_width)

def overlay_bbox(img, bbox, bbox_color=(0, 0, 255)):
    bbox_thickness = 2

    if len(bbox) == 8:
        bbox = np.round(np.array(bbox))
        pts = bbox.reshape((-1, 1, 2)).astype(np.int32)
        cv2.polylines(img, [pts], True, bbox_color, thickness=bbox_thickness, lineType=cv2.LINE_AA)
    elif len(bbox) == 4:
        tl_ = (int(round(bbox[0])), int(round(bbox[1])))
        br_ = (int(round(bbox[0] + bbox[2])), int(round(bbox[1] + bbox[3])))
        cv2.rectangle(img, tl_, br_, bbox_color, thickness=bbox_thickness)
    else:
        print('Error: Unknown tracker output.')
        exit(-1)

def overlay_circle(img, center, radius, circle_color=(0, 0, 255)):
    c = (int(round(center[1])), int(round(center[0])))
    cv2.circle(img, c, radius, circle_color, thickness=-1)


def run_tracker(sequence=None):
    visualize = True
    wait_timeout = 10

    dataset = VOTDataset('/home/benjamin/Desktop/D3S/sequences')

    if sequence is not None:
        dataset = [dataset[sequence]]

    tracker = Segm(params.parameters())

    for seq in dataset:

        print('--------------------------------------------')
        print('Tracking on sequence:', seq.name)

        img = cv2.cvtColor(cv2.imread(seq.frame(0).filename()), cv2.COLOR_BGR2RGB)

        m = seq.groundtruth(0).rasterize((0, 0, img.shape[1] - 1, img.shape[0] - 1)).astype(np.uint8)
        print(m)

        tracker.initialize(img, [], init_mask=m)

        if tracker.pred_mask is not None:
            overlay_mask(img, tracker.pred_mask)

        img = cv2.cvtColor(img, cv2.COLOR_RGB2BGR)

        if visualize:
            cv2.namedWindow('win', cv2.WINDOW_AUTOSIZE)
            cv2.imshow('win', img)
            #cv2.waitKey(0)

        times = []

        baseline_resp = [[], [], []] # 0 - tracking, 1 - not tracking (out of frame), 2 - not tracking (in frame)
        psr_resp = [[], [], []]
        psr_norm = [[], [], []]
        avgQT_over_QT = [[], [], []]

        for i in range(1, len(seq)):

            img = cv2.cvtColor(cv2.imread(seq.frame(i).filename()), cv2.COLOR_BGR2RGB)

            start_t = time.time()
            prediction = tracker.track(img)
            times.append(time.time() - start_t)
            p = Rectangle(prediction[0],prediction[1],prediction[2],prediction[3])
            gt = seq.groundtruth(i)
            out_of_frame = 1
            if isinstance(gt, Rectangle):
                print("In frame")
                if (calculate_overlap(p,gt) > 0.01):
                    out_of_frame = 0
                else:
                    out_of_frame = 2
            
            baseline_resp[out_of_frame].append(tracker.uncert_score)
            psr_resp[out_of_frame].append(tracker.current_psr)
            psr_norm[out_of_frame].append(tracker.current_psr_norm)
            avgQT_over_QT[out_of_frame].append(tracker.current_qt_avg)

            if visualize:
                img = copy.deepcopy(img)
                if tracker.pred_mask is not None:
                    overlay_mask(img, tracker.pred_mask, a=0.2, contour_line_width=1)

                # overlay_bbox(img, [gt.x, gt.y, gt.width, gt.height], bbox_color=(255, 0, 0))
                overlay_bbox(img, prediction)
                overlay_bbox(img, tracker.search_region_, bbox_color=(0, 0, 0))

                # if hasattr(tracker, 'reg_bb'):
                    # overlay_bbox(img, tracker.reg_bb, bbox_color=(255, 255, 0))

                if tracker.atom_pos is not None:
                    overlay_circle(img, tracker.atom_pos.numpy(), 5, (0, 0, 0))
                    overlay_circle(img, tracker.atom_pos.numpy(), 3, (255, 255, 255))

                pos = (10,25)
                fontColor = (0,255,0) if tracker.mdnet_score > 0 else (255,0,0)
                cv2.putText(img, "MDNet:" + str(round(tracker.mdnet_score*100)/100), 
                    pos, cv2.FONT_HERSHEY_SIMPLEX, 0.7,fontColor,2,2)

                if tracker.conflict:
                    cv2.putText(img, "CONFLICT FOUND", 
                    (10,65), cv2.FONT_HERSHEY_SIMPLEX, 1.2,(255,0,0),4,2)
                    overlay_bbox(img, tracker.conflictBox, bbox_color=(255, 255, 0))

                img = cv2.cvtColor(img, cv2.COLOR_RGB2BGR)

                cv2.imshow('win', img)
                key_ = cv2.waitKey(1500 if tracker.conflict else wait_timeout)

                if key_ == 27:
                    exit(10)
                elif key_ == 32:
                    if wait_timeout > 0:
                        wait_timeout = 0
                    else:
                        wait_timeout = 10

        print('--------------------------------')
        time_for_frame = sum(times) / len(times)
        print('Average speed: %.1f FPS' % (1 / time_for_frame))
        print('Milliseconds per-frame: %.1f' % (time_for_frame * 1000))
        print('--------------------------------')
        # print("Score averages when TRACKING target in frame")
        # print("Baseline: {0:.2f} \t PSR: {1:.2f} \t Normalized PSR: {2:.2f} \t avg(qt)/qt: {3:.2f}".format(np.mean(baseline_resp[0]), np.mean(psr_resp[0]), np.mean(psr_norm[0]), np.mean(avgQT_over_QT[0])))
        # print("STD")
        # print("Baseline: {0:.2f} \t PSR: {1:.2f} \t Normalized PSR: {2:.2f} \t avg(qt)/qt: {3:.2f}".format(np.std(baseline_resp[0]), np.std(psr_resp[0]), np.std(psr_norm[0]), np.std(avgQT_over_QT[0])))
        # print('--------------------------------')
        # print("Score averages when TRACKING target NOT in frame")
        # print("Baseline: {0:.2f} \t PSR: {1:.2f} \t Normalized PSR: {2:.2f} \t avg(qt)/qt: {3:.2f}".format(np.mean(baseline_resp[1]), np.mean(psr_resp[1]), np.mean(psr_norm[1]), np.mean(avgQT_over_QT[1])))
        # print("STD")
        # print("Baseline: {0:.2f} \t PSR: {1:.2f} \t Normalized PSR: {2:.2f} \t avg(qt)/qt: {3:.2f}".format(np.std(baseline_resp[1]), np.std(psr_resp[1]), np.std(psr_norm[1]), np.std(avgQT_over_QT[1])))
        # print('--------------------------------')
        # print("Score averages when NOT TRACKING target in frame")
        # print("Baseline: {0:.2f} \t PSR: {1:.2f} \t Normalized PSR: {2:.2f} \t avg(qt)/qt: {3:.2f}".format(np.mean(baseline_resp[2]), np.mean(psr_resp[2]), np.mean(psr_norm[2]), np.mean(avgQT_over_QT[2])))
        # print("STD")
        # print("Baseline: {0:.2f} \t PSR: {1:.2f} \t Normalized PSR: {2:.2f} \t avg(qt)/qt: {3:.2f}".format(np.std(baseline_resp[2]), np.std(psr_resp[2]), np.std(psr_norm[2]), np.std(avgQT_over_QT[2])))

def main():
    parser = argparse.ArgumentParser(description='Run tracker on sequence or dataset.')
    parser.add_argument('--sequence', type=str, default=None, help='Sequence number or name.')

    args = parser.parse_args()

    run_tracker(args.sequence)


if __name__ == '__main__':
    main()
