import os
import sys
import argparse
import time

import cv2
import numpy as np

env_path = os.path.join(os.path.dirname(__file__), '..')
if env_path not in sys.path:
    sys.path.append(env_path)

from vot.dataset import VOTDataset
from vot.region.shapes import Rectangle
from vot.region.raster import calculate_overlap

from pytracking.tracker.segm import Segm
from pytracking.parameter.segm import default_params as params


def draw_text(img, text, text_pos):
    font = cv2.FONT_HERSHEY_PLAIN
    text_sz = cv2.getTextSize(text, font, 1, 1)
    tl_ = (text_pos[0] - 5, text_pos[1] + 5)
    br_ = (text_pos[0] - 5 + text_sz[0][0] + 10, text_pos[1] - 5 - text_sz[0][1])
    cv2.rectangle(img, tl_, br_, (0, 0, 0), cv2.FILLED)
    cv2.putText(img, text, text_pos, font, 1, (255, 255, 255), 1, cv2.LINE_AA, False)

def draw_progress_bar(img, frame_index, sequence_length):
    cv2.line(img, (0, img.shape[0]), (img.shape[1], img.shape[0]), (0, 0, 0), 4)
    end_point = int(round((frame_index / sequence_length) * img.shape[1]))
    cv2.line(img, (0, img.shape[0]), (end_point, img.shape[0]), (255, 255, 255), 4)

def overlay_mask(img, mask, a=0.4, mask_color=(0, 255, 0), contour_line_width=2):
    if a >= 0:
        img_b = img[:, :, 0]
        img_g = img[:, :, 1]
        img_r = img[:, :, 2]

        img_r[mask > 0] = np.round(img_r[mask > 0] * (1 - a) + a * mask_color[0])
        img_g[mask > 0] = np.round(img_g[mask > 0] * (1 - a) + a * mask_color[1])
        img_b[mask > 0] = np.round(img_b[mask > 0] * (1 - a) + a * mask_color[2])

    M = (mask > 0).astype(np.uint8)
    if cv2.__version__[-5] == '4':
        contours, _ = cv2.findContours(M, cv2.RETR_EXTERNAL, cv2.CHAIN_APPROX_NONE)
    else:
        _, contours, _ = cv2.findContours(M, cv2.RETR_EXTERNAL, cv2.CHAIN_APPROX_NONE)

    cv2.drawContours(img, contours, -1, mask_color[::-1], contour_line_width)

def overlay_bbox(img, bbox, bbox_color=(0, 0, 255)):
    bbox_thickness = 2

    if len(bbox) == 8:
        bbox = np.round(np.array(bbox))
        pts = bbox.reshape((-1, 1, 2)).astype(np.int32)
        cv2.polylines(img, [pts], True, bbox_color, thickness=bbox_thickness, lineType=cv2.LINE_AA)
    elif len(bbox) == 4:
        tl_ = (int(round(bbox[0])), int(round(bbox[1])))
        br_ = (int(round(bbox[0] + bbox[2])), int(round(bbox[1] + bbox[3])))
        cv2.rectangle(img, tl_, br_, bbox_color, thickness=bbox_thickness)
    else:
        print('Error: Unknown tracker output.')
        exit(-1)

def overlay_circle(img, center, radius, circle_color=(0, 0, 255)):
    c = (int(round(center[1])), int(round(center[0])))
    cv2.circle(img, c, radius, circle_color, thickness=-1)


def rect_to_poly(rect):
    x0 = rect[0]
    y0 = rect[1]
    x1 = rect[0] + rect[2]
    y1 = rect[1]
    x2 = rect[0] + rect[2]
    y2 = rect[1] + rect[3]
    x3 = rect[0]
    y3 = rect[1] + rect[3]
    return [x0, y0, x1, y1, x2, y2, x3, y3]

def mask_from_region(region, image_sz):
    if len(region) == 4:
        region = rect_to_poly(region)
    mask = np.zeros((image_sz[1], image_sz[0]), dtype=np.uint8)
    x_ = [region[i] for i in range(0, len(region), 2)]
    y_ = [region[i] for i in range(1, len(region), 2)]
    x0 = max(0, round(min(x_)))
    y0 = max(0, round(min(y_)))
    x1 = min(image_sz[0] - 1, round(max(x_)))
    y1 = min(image_sz[1] - 1, round(max(y_)))
    mask[y0:y1+1, x0:x1+1] = 1
    return mask

def getPredictionList(path):
    f1 = open(path, "r")
    lines = []
    f1.readline()
    for l in f1.readlines():
        parts = l.strip().split(",")
        lines.append((int(parts[0]) , int(parts[1]), int(parts[2]), int(parts[3])))
    return lines

def getConfidences(path):
    f1 = open(path, "r")
    lines = []
    f1.readline()
    for l in f1.readlines():
        lines.append(float(l.strip()))
      
    return lines

def run_tracker(sequence=None):
    visualize = True
    wait_timeout = 0

    dataset = VOTDataset('/home/benjamin/Desktop/D3S/sequences') #'C:\\Users\\Beni\\Desktop\\D3S\\d3s\\VOT2020ST'

    if sequence is not None:
        dataset = [dataset[sequence]]

    predpath = '/home/benjamin/Desktop/D3S/results/D3SLT/longterm/' + sequence + '/' + sequence + '_001.txt'
    preds = getPredictionList(predpath)

    valspath = '/home/benjamin/Desktop/D3S/results/D3SLT/longterm/' + sequence + '/' + sequence + '_001_confidence.value'
    vals = getConfidences(valspath)
    # tracker = Segm(params.parameters())
    for seq in dataset:

        print('--------------------------------------------')
        print('Tracking on sequence:', seq.name)

        img = cv2.cvtColor(cv2.imread(seq.frame(0).filename()), cv2.COLOR_BGR2RGB)
        m = seq.groundtruth(0).rasterize((0, 0, img.shape[1] - 1, img.shape[0] - 1)).astype(np.uint8)

        img = cv2.cvtColor(img, cv2.COLOR_RGB2BGR)

        if visualize:
            cv2.namedWindow('win', cv2.WINDOW_AUTOSIZE)
            cv2.imshow('win', img)
            #cv2.waitKey(0)

        times = []
        print(len(seq))
        print(len(preds))
        print(len(vals))

        for i in range(1, len(seq)):

            img = cv2.cvtColor(cv2.imread(seq.frame(i).filename()), cv2.COLOR_BGR2RGB)

            start_t = time.time()
            prediction = preds[i-1] 
            times.append(time.time() - start_t)
            p = Rectangle(prediction[0],prediction[1],prediction[2],prediction[3])
            gt = seq.groundtruth(i)
            out_of_frame = 1
            if isinstance(gt, Rectangle):
                if prediction[0] < 2*img.shape[0]:
                    if (calculate_overlap(p,gt) > 0.01):
                        out_of_frame = 0
                    else:
                        out_of_frame = 2


            if visualize:
                # if tracker.pred_mask is not None:
                #     overlay_mask(img, tracker.pred_mask, a=0.2, contour_line_width=1)

                if out_of_frame != 1:
                    overlay_bbox(img, [gt.x, gt.y, gt.width, gt.height], bbox_color=(255, 0, 0))
                
                # weird bug
                if prediction[0] < 2*img.shape[0]:
                    overlay_bbox(img, prediction)
                # overlay_bbox(img, tracker.search_region_, bbox_color=(0, 0, 0))

                # if hasattr(tracker, 'reg_bb'):
                    # overlay_bbox(img, tracker.reg_bb, bbox_color=(255, 255, 0))

                # if tracker.atom_pos is not None:
                #     overlay_circle(img, tracker.atom_pos.numpy(), 5, (0, 0, 0))
                #     overlay_circle(img, tracker.atom_pos.numpy(), 3, (255, 255, 255))

                img = cv2.cvtColor(img, cv2.COLOR_RGB2BGR)


                font = cv2.FONT_HERSHEY_SIMPLEX
                pos = (10,25)
                fontScale = 1
                fontColor =  (0,255,0) if 1 / vals[i-1] > 0 else (0,0,255)
                thickness = 2
                lineType = 2

                cv2.putText(img, str(round(vals[i-1]*100)/100), 
                    pos, 
                    font, 
                    fontScale,
                    fontColor,
                    thickness,
                    lineType)

                cv2.imshow('win', img)
                key_ = cv2.waitKey(wait_timeout)

                if key_ == 27:
                    exit(10)
                elif key_ == 32:
                    if wait_timeout > 0:
                        wait_timeout = 0
                    else:
                        wait_timeout = 10


def main():
    parser = argparse.ArgumentParser(description='Run tracker on sequence or dataset.')
    parser.add_argument('--sequence', type=str, default=None, help='Sequence number or name.')

    args = parser.parse_args()

    run_tracker(args.sequence)


if __name__ == '__main__':
    main()
