from tabnanny import check
from pytracking.tracker.base import BaseTracker
import torch
from random import choices, seed
import torch.nn.functional as F
import torch.nn
import math
import time
import numpy as np
import cv2
import copy
from skimage.feature import peak_local_max
from pytracking import dcf, fourier, TensorList, operation
from pytracking.features.preprocessing import numpy_to_torch
from pytracking.utils.plotting import show_tensor
from pytracking.libs.optimization import GaussNewtonCG, ConjugateGradient, GradientDescentL2
from pytracking.parameter.segm import default_params as params
from .optim import ConvProblem, FactorizedConvProblem
from pytracking.features import augmentation
import ltr.data.processing_utils as prutils
from ltr import load_network
from external.RTMDNet.tracker import get_score_mdnet, init_mdnet, update_mdnet, generate_samples_neg

from pytracking.bbox_fit import fit_bbox_to_mask, segm_fit_aa_box
from pytracking.mask_to_disk import save_mask

torch.backends.cudnn.benchmark = False
torch.backends.cudnn.deterministic = True


class Point:
    """
    This class generate points.
    """
    def __init__(self, stride, size, image_center):
        self.stride = stride
        self.size = size
        self.image_center = image_center

        self.points = self.generate_points(self.stride, self.size, self.image_center)

    def generate_points(self, stride, size, im_c):
        ori = im_c - size // 2 * stride
        x, y = np.meshgrid([ori + stride * dx for dx in np.arange(0, size)],
                           [ori + stride * dy for dy in np.arange(0, size)])
        points = np.zeros((2, size, size), dtype=np.float32)
        points[0, :, :], points[1, :, :] = x.astype(np.float32), y.astype(np.float32)

        return points

def calc_change(r):
    return np.maximum(r, 1. / r)

def calc_sz(w, h):
    pad = (w + h) * 0.5
    return np.sqrt((w + pad) * (h + pad))

class Segm(BaseTracker):

    def initialize_features(self): ### Initializes the backbone
        if not getattr(self, 'features_initialized', False):
            self.params.features_filter.initialize()
        self.features_initialized = True

    def _rect_from_mask(self, mask, min_max=False):
        '''
        create an axis-aligned rectangle from a given binary mask
        mask in created as a minimal rectangle containing all non-zero pixels
        '''
        return fit_bbox_to_mask(mask.astype(np.int32), rotated=True, alpha=1)

    def segm_rect_from_mask(self, mask):
        target_mask = (mask > 0.5).astype(np.float32)
        #aa_box = segm_fit_aa_box(target_mask, 0.1, rotated=True, alpha=0.1)
        aa_box = segm_fit_aa_box(target_mask, 0.1, rotated=True, alpha=0.1)
        return np.array([[aa_box[0], aa_box[1]], [aa_box[2], aa_box[1]],
                         [aa_box[2], aa_box[3]], [aa_box[0], aa_box[3]]])
    
    def initialize_short_term(self, image, state, *args, **kwargs):
        torch.manual_seed(463)
        seed(2)
        self.frame_num = 1

        if not hasattr(self.params, 'device'):
            self.params.device = 'cuda' if self.params.use_gpu else 'cpu'

        self.pred_mask = None
        self.atom_pos = None

        # Initialize features
        self.initialize_features()

        # Check if image is color
        self.params.features_filter.set_is_color(image.shape[2] == 3)

        # Get feature specific params
        self.fparams = self.params.features_filter.get_fparams('feature_params')
        
        self.rotated_bbox = True
        self.report_aa = False
        if len(state) == 4:
            state[0] -= 1
            state[1] -= 1
            # Get position and size
            self.pos = torch.Tensor([state[1] + state[3] / 2, state[0] + state[2] / 2])
            self.pos_prev = [state[1] + state[3] / 2, state[0] + state[2] / 2]
            self.target_sz = torch.Tensor([state[3], state[2]])
            self.gt_poly = np.array([state[0], state[1],
                                        state[0] + state[2] - 1, state[1],
                                        state[0] + state[2] - 1, state[1] + state[3] - 1,
                                        state[0], state[1] + state[3] - 1])

            self.rotated_bbox = False
            self.report_aa = True

        self.bb_sz = np.array([self.target_sz[1].item(), self.target_sz[0].item()])

        if self.params.bbox_reg_level == 3:
            hanning = np.hanning(16)
        elif self.params.bbox_reg_level == 2:
            hanning = np.hanning(36)
        self.bb_reg_window = np.outer(hanning, hanning).flatten()

        target_image_size_ratio = (self.target_sz[0] * self.target_sz[1]).item() / (image.shape[0] * image.shape[1])
        if target_image_size_ratio < 0.005:
            self.params.max_image_sample_size = (22 * 16) ** 2
            self.params.min_image_sample_size = (22 * 16) ** 2
            self.params.search_area_scale = 6

        # Set search area
        self.target_scale = 1.0
        search_area = torch.prod(self.target_sz * self.params.search_area_scale).item()
        if search_area > self.params.max_image_sample_size:
            self.target_scale = math.sqrt(search_area / self.params.max_image_sample_size)
        elif search_area < self.params.min_image_sample_size:
            self.target_scale = math.sqrt(search_area / self.params.min_image_sample_size)

        self.target_scale = torch.Tensor([self.target_scale])

        # Target size in base scale
        self.base_target_sz = self.target_sz / self.target_scale

        # Use odd square search area and set sizes
        feat_max_stride = max(self.params.features_filter.stride())
        if getattr(self.params, 'search_area_shape', 'square') == 'square':
            self.img_sample_sz = torch.round(
                torch.sqrt(torch.prod(self.base_target_sz * self.params.search_area_scale))) * torch.ones(2)
        elif self.params.search_area_shape == 'initrect':
            self.img_sample_sz = torch.round(self.base_target_sz * self.params.search_area_scale)
        else:
            raise ValueError('Unknown search area shape')
        if self.params.feature_size_odd:
            self.img_sample_sz += feat_max_stride - self.img_sample_sz % (2 * feat_max_stride)
        else:
            self.img_sample_sz += feat_max_stride - (self.img_sample_sz + feat_max_stride) % (2 * feat_max_stride)

        # Set sizes
        self.img_support_sz = self.img_sample_sz
        self.feature_sz = self.params.features_filter.size(self.img_sample_sz)
        self.output_sz = self.params.score_upsample_factor * self.img_support_sz  # Interpolated size of the output
        self.kernel_size = self.fparams.attribute('kernel_size')

        self.global_sz = None # size of scaled whole image in pixels before feature extraction
        self.global_feat_sz = None # size of features of whole image, can change throughout tracking proportionally to scale

        # Optimization options
        self.params.precond_learning_rate = self.fparams.attribute('learning_rate')
        if self.params.CG_forgetting_rate is None or max(self.params.precond_learning_rate) >= 1:
            self.params.direction_forget_factor = 0
        else:
            self.params.direction_forget_factor = (1 - max(
                self.params.precond_learning_rate)) ** self.params.CG_forgetting_rate

        self.output_window = None
        if getattr(self.params, 'window_output', False):
            if getattr(self.params, 'use_clipped_window', False):
                self.output_window = dcf.hann2d_clipped(self.output_sz.long(),
                                                        self.output_sz.long() * self.params.effective_search_area / self.params.search_area_scale,
                                                        centered=False).to(self.params.device)
            else:
                self.output_window = dcf.hann2d(self.output_sz.long(), centered=False).to(self.params.device)

        # Initialize some learning things
        self.init_learning()

        # Convert image
        im = numpy_to_torch(image)
        self.im = im  # For debugging only

        # Setup scale bounds
        self.image_sz = torch.Tensor([im.shape[2], im.shape[3]])
        self.min_scale_factor = torch.max(10 / self.base_target_sz)
        self.max_scale_factor = torch.min(self.image_sz / self.base_target_sz)

        # Extract and transform sample
        x = self.generate_init_samples(im)

        # Initialize projection matrix
        self.init_projection_matrix(x)

        # Transform to get the training sample
        train_x = self.preprocess_sample(x)

        # Generate label function
        init_y = self.init_label_function(train_x)

        # Init memory
        self.init_memory(train_x)

        # Init optimizer and do initial optimization .. DCF?
        self.init_optimization(train_x, init_y)

        self.state_tracking = state

        self.init_segmentation(image, state, init_mask=None)
        print("Initialized.")

    def initialize(self, image, state, init_mask=None, *args, **kwargs):

        torch.manual_seed(466)
        seed(6)
        # Initialize some stuff
        self.frame_num = 1

        if not hasattr(self.params, 'device'):
            self.params.device = 'cuda' if self.params.use_gpu else 'cpu'

        self.pred_mask = None
        self.atom_pos = None

        # Initialize features
        self.initialize_features()

        # Check if image is color
        self.params.features_filter.set_is_color(image.shape[2] == 3)

        # Get feature specific params
        self.fparams = self.params.features_filter.get_fparams('feature_params')

        self.time = 0
        tic = time.time()

        self.rotated_bbox = True
        self.report_aa = False
        if init_mask is not None:
            if len(state) == 0:
                if self.params.tracking:
                    rect_ = self._rect_from_mask(init_mask) ### Using "old" bbox fitting? 
                    if rect_ is None:
                        rect_ = self.segm_rect_from_mask(init_mask)
                else:
                    rect_ = self.segm_rect_from_mask(init_mask)

                self.pos = torch.Tensor([np.mean(rect_[:, 1]), np.mean(rect_[:, 0])])
                self.pos_prev = [np.mean(rect_[:, 1]), np.mean(rect_[:, 0])]
                self.gt_poly = rect_.flatten()

                state = self.poly_to_aabbox(rect_[:, 0], rect_[:, 1])
            elif len(state) == 4:
                self.pos = torch.Tensor([state[1] + state[3] / 2, state[0] + state[2] / 2])
                self.pos_prev = [state[1] + state[3] / 2, state[0] + state[2] / 2]
                self.gt_poly = np.array([state[0], state[1],
                                         state[0] + state[2], state[1],
                                         state[0] + state[2], state[1] + state[3],
                                         state[0], state[1] + state[3]])
            else:
                print('This is not implemented yet.')
                exit(-1)
            self.target_sz = torch.Tensor([state[3], state[2]])
            self.rotated_bbox = False
            self.mask_prediction = None
        else:
            if len(state) == 4:
                state[0] -= 1
                state[1] -= 1
                # Get position and size
                self.pos = torch.Tensor([state[1] + state[3] / 2, state[0] + state[2] / 2])
                self.pos_prev = [state[1] + state[3] / 2, state[0] + state[2] / 2]
                self.target_sz = torch.Tensor([state[3], state[2]])
                self.gt_poly = np.array([state[0], state[1],
                                         state[0] + state[2] - 1, state[1],
                                         state[0] + state[2] - 1, state[1] + state[3] - 1,
                                         state[0], state[1] + state[3] - 1])

                self.rotated_bbox = False
                self.report_aa = True
            else:
                raise NotImplementedError('Only 4-element region initialization supported.')

        # for bounding box regression:
        self.bb_sz = np.array([self.target_sz[1].item(), self.target_sz[0].item()])

        if self.params.bbox_reg_level == 3:
            hanning = np.hanning(16)
        elif self.params.bbox_reg_level == 2:
            hanning = np.hanning(36)
        self.bb_reg_window = np.outer(hanning, hanning).flatten()

        self.last_target_sz = [self.target_sz[1].item(), self.target_sz[0].item()]

        target_image_size_ratio = (self.target_sz[0] * self.target_sz[1]).item() / (image.shape[0] * image.shape[1])
        if target_image_size_ratio < 0.005:
            self.params.max_image_sample_size = (22 * 16) ** 2
            self.params.min_image_sample_size = (22 * 16) ** 2
            self.params.search_area_scale = 6


        # Set search area
        self.target_scale = 1.0
        search_area = torch.prod(self.target_sz * self.params.search_area_scale).item()
        if search_area > self.params.max_image_sample_size:
            self.target_scale = math.sqrt(search_area / self.params.max_image_sample_size)
        elif search_area < self.params.min_image_sample_size:
            self.target_scale = math.sqrt(search_area / self.params.min_image_sample_size)

        self.target_scale = torch.Tensor([self.target_scale])
        self.init_target_scale = copy.deepcopy(self.target_scale)

        # Target size in base scale
        self.base_target_sz = self.target_sz / self.target_scale

        # Use odd square search area and set sizes
        feat_max_stride = max(self.params.features_filter.stride())
        if getattr(self.params, 'search_area_shape', 'square') == 'square':
            self.img_sample_sz = torch.round(
                torch.sqrt(torch.prod(self.base_target_sz * self.params.search_area_scale))) * torch.ones(2)
        elif self.params.search_area_shape == 'initrect':
            self.img_sample_sz = torch.round(self.base_target_sz * self.params.search_area_scale)
        else:
            raise ValueError('Unknown search area shape')
        if self.params.feature_size_odd:
            self.img_sample_sz += feat_max_stride - self.img_sample_sz % (2 * feat_max_stride)
        else:
            self.img_sample_sz += feat_max_stride - (self.img_sample_sz + feat_max_stride) % (2 * feat_max_stride)

        # Set sizes
        self.img_support_sz = self.img_sample_sz
        self.feature_sz = self.params.features_filter.size(self.img_sample_sz)
        self.output_sz = self.params.score_upsample_factor * self.img_support_sz  # Interpolated size of the output
        self.kernel_size = self.fparams.attribute('kernel_size')

        self.global_sz = None # size of scaled whole image in pixels before feature extraction
        self.global_feat_sz = None # size of features of whole image, can change throughout tracking proportionally to scale

        # Optimization options
        self.params.precond_learning_rate = self.fparams.attribute('learning_rate')
        if self.params.CG_forgetting_rate is None or max(self.params.precond_learning_rate) >= 1:
            self.params.direction_forget_factor = 0
        else:
            self.params.direction_forget_factor = (1 - max(
                self.params.precond_learning_rate)) ** self.params.CG_forgetting_rate

        self.output_window = None
        if getattr(self.params, 'window_output', False):
            if getattr(self.params, 'use_clipped_window', False):
                self.output_window = dcf.hann2d_clipped(self.output_sz.long(),
                                                        self.output_sz.long() * self.params.effective_search_area / self.params.search_area_scale,
                                                        centered=False).to(self.params.device)
            else:
                self.output_window = dcf.hann2d(self.output_sz.long(), centered=False).to(self.params.device)

        # Initialize some learning things
        self.init_learning()

        # Convert image
        im = numpy_to_torch(image)
        self.im = im  # For debugging only

        # Setup scale bounds
        self.image_sz = torch.Tensor([im.shape[2], im.shape[3]])
        self.min_scale_factor = torch.max(10 / self.base_target_sz)
        self.max_scale_factor = torch.min(self.image_sz / self.base_target_sz)

        # Extract and transform sample
        x = self.generate_init_samples(im)

        # Initialize projection matrix
        self.init_projection_matrix(x)

        # Transform to get the training sample
        train_x = self.preprocess_sample(x)

        # Generate label function
        init_y = self.init_label_function(train_x)

        # Init memory
        self.init_memory(train_x)

        # Init optimizer and do initial optimization .. DCF?
        self.init_optimization(train_x, init_y)

        self.init_filter = copy.deepcopy(self.filter)

        if self.params.use_segmentation:
            self.init_segmentation(image, state, init_mask=init_mask)

        # array of scores
        self.scores = np.array([1])

        self.scores2 = np.array([1]) # PSR-based scores
        self.psr_norm = 0 # First frame PSR

        self.redetection_mode = 0 #should perform redetection?
        self.undetected_frames = 0 # How many frames has redetection been turned on
        self.last_good_position = torch.tensor([state[1]+state[3]/2, state[0]+state[2]/2])
        self.succesfuly_tracked_frames = 0 # How many frames in a row have been succesfully tracked (no redetection)
        self.mdnet_score = 0

        self.current_psr = 0 # For analysis
        self.current_psr_norm = 0
        self.current_qt_avg = 0

        self.model, self.cropper, self.mdnet_criterion, self.mdnet_optimizer, self.positive_feats, self.negative_feats = init_mdnet(image, state)
        self.init_model = copy.deepcopy(self.model)

        self.state_tracking = state

        self.numKeptFrames = 260
        self.img_buffer = [image] # keeps last 100 frames, newest frame at beginning.
        self.state_buffer = [(self.frame_num,copy.deepcopy(self.state_tracking))] # keeps last 100 succesful track states, (frame_num, state) newest state at beginning.
        self.temp_state_buffer = [] # eventually gets merged into state_buffer
        self.shouldBacktrack = False # boolean indicator which changes throughout tracking under certain conditions
        self.alreadyBacktracked = False # flag which is used to not unnecessarily re-run backtrack on a target which is continued to be tracked confidently.
        self.conflict = False
        self.conflictBox = None
        toc_ = time.time() - tic
        self.time += toc_

    def init_optimization(self, train_x, init_y):
        # Initialize filter
        filter_init_method = getattr(self.params, 'filter_init_method', 'zeros')
        self.filter = TensorList(
            [x.new_zeros(1, cdim, sz[0], sz[1]) for x, cdim, sz in zip(train_x, self.compressed_dim, self.kernel_size)])
        if filter_init_method == 'zeros':
            pass
        elif filter_init_method == 'randn':
            for f in self.filter:
                f.normal_(0, 1 / f.numel())
        else:
            raise ValueError('Unknown "filter_init_method"')

        # Get parameters
        self.params.update_projection_matrix = getattr(self.params, 'update_projection_matrix',
                                                       True) and self.params.use_projection_matrix
        optimizer = getattr(self.params, 'optimizer', 'GaussNewtonCG')

        # Setup factorized joint optimization
        if self.params.update_projection_matrix:
            self.joint_problem = FactorizedConvProblem(self.init_training_samples, init_y, self.filter_reg,
                                                       self.fparams.attribute('projection_reg'), self.params,
                                                       self.init_sample_weights,
                                                       self.projection_activation, self.response_activation)

            # Variable containing both filter and projection matrix
            joint_var = self.filter.concat(self.projection_matrix)

            # Initialize optimizer
            analyze_convergence = getattr(self.params, 'analyze_convergence', False)
            if optimizer == 'GaussNewtonCG':
                self.joint_optimizer = GaussNewtonCG(self.joint_problem, joint_var, plotting=(self.params.debug >= 3),
                                                     analyze=analyze_convergence, fig_num=(12, 13, 14))
            elif optimizer == 'GradientDescentL2':
                self.joint_optimizer = GradientDescentL2(self.joint_problem, joint_var,
                                                         self.params.optimizer_step_length,
                                                         self.params.optimizer_momentum,
                                                         plotting=(self.params.debug >= 3), debug=analyze_convergence,
                                                         fig_num=(12, 13))

            # Do joint optimization
            if isinstance(self.params.init_CG_iter, (list, tuple)):
                self.joint_optimizer.run(self.params.init_CG_iter)
            else:
                self.joint_optimizer.run(self.params.init_CG_iter // self.params.init_GN_iter, self.params.init_GN_iter)

            if analyze_convergence:
                opt_name = 'CG' if getattr(self.params, 'CG_optimizer', True) else 'GD'
                for val_name, values in zip(['loss', 'gradient'],
                                            [self.joint_optimizer.losses, self.joint_optimizer.gradient_mags]):
                    val_str = ' '.join(['{:.8e}'.format(v.item()) for v in values])
                    file_name = '{}_{}.txt'.format(opt_name, val_name)
                    with open(file_name, 'a') as f:
                        f.write(val_str + '\n')
                raise RuntimeError('Exiting')

        # Re-project samples with the new projection matrix
        compressed_samples = self.project_sample(self.init_training_samples, self.projection_matrix)
        for train_samp, init_samp in zip(self.training_samples, compressed_samples):
            train_samp[:init_samp.shape[0], ...] = init_samp

        self.hinge_mask = None

        # Initialize optimizer
        self.conv_problem = ConvProblem(self.training_samples, self.y, self.filter_reg, self.sample_weights,
                                        self.response_activation)

        if optimizer == 'GaussNewtonCG':
            self.filter_optimizer = ConjugateGradient(self.conv_problem, self.filter,
                                                      fletcher_reeves=self.params.fletcher_reeves,
                                                      direction_forget_factor=self.params.direction_forget_factor,
                                                      debug=(self.params.debug >= 3), fig_num=(12, 13))
        elif optimizer == 'GradientDescentL2':
            self.filter_optimizer = GradientDescentL2(self.conv_problem, self.filter, self.params.optimizer_step_length,
                                                      self.params.optimizer_momentum, debug=(self.params.debug >= 3),
                                                      fig_num=12)

        # Transfer losses from previous optimization
        if self.params.update_projection_matrix:
            self.filter_optimizer.residuals = self.joint_optimizer.residuals
            self.filter_optimizer.losses = self.joint_optimizer.losses

        if not self.params.update_projection_matrix:
            self.filter_optimizer.run(self.params.init_CG_iter)

        # Post optimization
        self.filter_optimizer.run(self.params.post_init_CG_iter)

        # Free memory
        del self.init_training_samples
        if self.params.use_projection_matrix:
            del self.joint_problem, self.joint_optimizer

    def track(self, image):
        print("#################################")
        self.frame_num += 1
        self.img_buffer.insert(0,image)
        if len(self.img_buffer) > self.numKeptFrames:
            self.img_buffer.pop()

        self.pred_mask = None
        self.atom_pos = None
        self.siamban_bbox = None

        self.max_1 = None
        self.max_2 = None
        self.translation_1 = None
        self.translation_2 = None

        self.frame_name = '%08d' % self.frame_num

        self.pos_prev = [copy.copy(self.pos[0].item()), copy.copy(self.pos[1].item())]

        # Convert image
        im = numpy_to_torch(image)
        self.im = im  # For debugging only

        startitme = time.time()

        # ------- LOCALIZATION ------- #
        sample_scales = self.target_scale * self.params.scale_factors # only one scale factor: this is the scale factor between search
        if self.undetected_frames > 3 and self.undetected_frames%3 == 0:
           sample_scales = self.init_target_scale * self.params.scale_factors
        sample_pos = None
        has_global = False
        ## Should do global redetection? 
        if self.redetection_mode == 1 and image.shape[0] <= 1080:
            # pass
            if sample_scales[0] < 0.2:
                sample_scales[0] = 0.21
            test2x = self.extract_processed_image(im, sample_scales, self.img_sample_sz)
            self.global_feat_sz = [torch.tensor(test2x[0].size()[-2:])] # global feature size
            scores_raw_full = self.apply_filter(test2x)
            new_pos_global, s_global = self.localize_target(scores_raw_full,globl=True)
            has_global = True
            sample_pos = copy.deepcopy(new_pos_global)
        else:
            # Get sample
            sample_pos = copy.deepcopy(self.pos)

        test_x = self.extract_processed_sample(im, sample_pos, sample_scales, self.img_sample_sz) # in: original image, center position of target + scale of search region, 
        # out: 1x256x14x14 features from ATOMResNet50 of search region
        time1 = time.time()
        search_region_sz_ = sample_scales[0] * self.img_sample_sz
        
        x0_ = int(round((sample_pos[1] - search_region_sz_[1] / 2).item()))
        y0_ = int(round((sample_pos[0] - search_region_sz_[0] / 2).item()))
        self.search_region_ = [x0_, y0_, int(round(search_region_sz_[1].item())), int(round(search_region_sz_[0].item()))]
        # Compute scores
        scores_raw = self.apply_filter(test_x) # in: 1x256x14x14 features from ATOMResNet50 of search region,# self.filter is 1x256x4x4
        # out: 1x1x14x14 features of search region convolved with learned DCF target template. #maybe i'll need to keep a scale=1 filter/template of the
        
        time2 = time.time()

        translation_vec, scale_ind, s, flag = self.localize_target(scores_raw) ### GEM  # s = 1x1x224x224 % upsampled from a 14x14 response  
        new_pos = sample_pos + translation_vec

        time3 = time.time()

        psr_curr = self.psr(s)
        if self.frame_num == 2:
            self.psr_norm = psr_curr

        # Localization uncertainty ### Some sort of absence detection already implemented?
        max_score = torch.max(s).item()
        uncert_score = 0
        qt_curr = psr_curr / self.psr_norm # Normalize by first frame value
        qt_avg = 1
        if self.frame_num > 5:
            uncert_score = np.mean(self.scores) / max_score # keep the reliable scores here
            qt_avg = np.mean(self.scores2)
        self.uncert_score = uncert_score

        # OLD SCORE-BASED RULE
        # if (uncert_score < self.params.tracking_uncertainty_thr and not self.redetection_mode) or uncert_score < self.params.tracking_uncertainty_thr/2:
        #     self.scores = np.append(self.scores, max_score)
        #     if self.scores.size > self.params.response_budget_sz: # score budget
        #         self.scores = np.delete(self.scores, 0)
        
        # NEW PSR BASED RULE
        # if qt_avg/qt_curr < 2.7:
        #     self.succesfuly_tracked_frames+=1
        #     if not self.redetection_mode:
        #         self.scores2 = np.append(self.scores2, qt_curr)
        #         if self.scores2.size > 100: # score budget
        #             self.scores2 = np.delete(self.scores2, 0)
        #     else:
        #         self.setRedetection(0)
        # else:
        #     self.setRedetection(1)
        #     uncert_score = 100
        #     self.succesfuly_tracked_frames = 0

        # Update position and scale
        if uncert_score < self.params.tracking_uncertainty_thr and self.succesfuly_tracked_frames > 3:
            if getattr(self.params, 'use_classifier', True):
                self.update_state(new_pos, sample_scales[scale_ind])
        # just a sanity check so that it does not get out of image
        if new_pos[0] < 0:
            new_pos[0] = 0
        if new_pos[1] < 0:
            new_pos[1] = 0
        if new_pos[0] >= image.shape[0]:
            new_pos[0] = image.shape[0] - 1
        if new_pos[1] >= image.shape[1]:
            new_pos[1] = image.shape[1] - 1

        size_target_tensor = torch.tensor([self.state_tracking[2], self.state_tracking[3]])
        proposed_bbox = torch.cat((new_pos[[1, 0]] - (size_target_tensor - 1) / 2, size_target_tensor)).tolist()
        ### GIM Forward pass
        self.atom_pos = new_pos
        pred_segm_region = None
        if self.segmentation_task or (
            self.params.use_segmentation and 0 < self.params.uncertainty_segment_thr): # only call GIM... if GEM is certain enough that the target is there
            pred_segm_region = self.segment_target(image, new_pos, self.target_sz)
            if pred_segm_region is None:
                self.pos = new_pos.clone()
            else:
                proposed_bbox = np.array([pred_segm_region[0], pred_segm_region[1], pred_segm_region[2], pred_segm_region[3]])
        else:
            self.pos = new_pos.clone()

        # NEWEST MDNet BASED RULE
        score, _ = get_score_mdnet(image, self.model, np.array(proposed_bbox) , self.cropper, proposed_bbox, additional_pos_samps=0)
        print("CENTER MASK (y,x):", self.state_tracking[1]+self.state_tracking[3]/2, self.state_tracking[0]+self.state_tracking[2]/2)
        self.mdnet_score = score
        print("SCORE:",score)
        if score > 0:
            self.succesfuly_tracked_frames+=1
            self.setRedetection(0)
        elif self.frame_num > 5:
            self.setRedetection(1)
            uncert_score = 100
            self.succesfuly_tracked_frames = 0
            self.temp_state_buffer = []
            self.alreadyBacktracked = False
        
        if flag == 'not_found':
            uncert_score = 100
            self.setRedetection(1)
            self.succesfuly_tracked_frames = 0
            self.temp_state_buffer = []
            self.alreadyBacktracked = False

        if self.redetection_mode:
            self.undetected_frames+=1
            if self.undetected_frames > 2:
                self.shouldBacktrack = True
        elif self.succesfuly_tracked_frames <= 3:
            self.undetected_frames+=1
        else:
            self.undetected_frames=0
            

        if not self.redetection_mode:
            if self.succesfuly_tracked_frames <= 3:
                self.temp_state_buffer.insert(0,(self.frame_num, copy.deepcopy(np.array(proposed_bbox))))
            else:
                self.last_good_position = copy.deepcopy(new_pos)
                self.state_buffer.insert(0,(self.frame_num, copy.deepcopy(np.array(proposed_bbox))))
                if len(self.state_buffer) > self.numKeptFrames:
                    self.state_buffer.pop()

        if self.frame_num > 5 and not self.redetection_mode and self.succesfuly_tracked_frames == 3 and self.shouldBacktrack and not self.alreadyBacktracked:
            isTP, feat_buffer = self.backtrack(False) # -1 low MDNet, 0 low motion model/no trajectory overlap, 1 TP
            self.shouldBacktrack = False
            # If this is the true target, just continue.. 
            if len(feat_buffer) > 0:
                catfeats = torch.cat(feat_buffer)
                if isTP == 1:
                    idx = choices(range(catfeats.size()[0]),k=300)
                    self.positive_feats.append(catfeats[idx,:])
                    self.state_buffer = self.temp_state_buffer + self.state_buffer
                    while len(self.state_buffer) > self.numKeptFrames:
                        self.state_buffer.pop()
                    bboxes_neg = generate_samples_neg(image, np.array(self.state_tracking), 40)
                    proposed_bbox = copy.deepcopy(self.state_tracking)
                    update_mdnet(self.model, image, np.array(proposed_bbox), None, np.array(bboxes_neg) if len(bboxes_neg) > 0 else None, self.mdnet_criterion, self.mdnet_optimizer, 
                            self.cropper, self.positive_feats, self.negative_feats, additional_neg_samples=0, additional_pos_samples=0)
                else:
                    if isTP == 0:
                        idx = choices(range(catfeats.size()[0]),k=1000)
                        self.negative_feats.append(catfeats[idx,:])
                        bboxes_neg = generate_samples_neg(image, np.array(self.state_tracking), 40)
                        proposed_bbox = copy.deepcopy(self.state_tracking)
                        update_mdnet(self.model, image, np.array(proposed_bbox), None, np.array(bboxes_neg) if len(bboxes_neg) > 0 else None, self.mdnet_criterion, self.mdnet_optimizer, 
                            self.cropper, self.positive_feats, self.negative_feats, additional_neg_samples=0, additional_pos_samples=0)
                        
                    # re-enter redetection
                    self.setRedetection(1)
                    self.alreadyBacktracked = True
                    uncert_score = 100
                    self.succesfuly_tracked_frames = 0
                    self.temp_state_buffer = []

        self.current_psr = psr_curr
        self.current_psr_norm = qt_curr
        self.current_qt_avg = qt_avg/qt_curr
        # print("Running average max response/Current max response: {0:.2f} \t PSR: {1:.2f}, n:{2:.2f} \t AVG qt: {3:.2f} \t qtA/qt: {4:.2f} ".format(self.uncert_score,  psr_curr, qt_curr, np.mean(self.scores2), qt_avg/qt_curr ))

        # self.params.debug = 2
        if self.params.debug >= 2:
            show_tensor(s[scale_ind, ...], 5, title='Max score = {:.2f}'.format(torch.max(s[scale_ind, ...]).item()))
            if (self.redetection_mode and has_global):
                show_tensor(s_global[scale_ind, ...], 6, title='Max score global = {:.2f}'.format(torch.max(s_global[scale_ind, ...]).item()))

        time4 = time.time()
        # ------- UPDATE ------- #

        # Check flags and set learning rate if hard negative
        update_flag = flag not in ['not_found', 'uncertain']
        hard_negative = (flag == 'hard_negative')
        learning_rate = self.params.hard_negative_learning_rate if hard_negative else None

        if update_flag and score > 0 and self.succesfuly_tracked_frames >3:
            # Get train sample
            train_x = TensorList([x[scale_ind:scale_ind + 1, ...] for x in test_x])

            # Create label for sample
            train_y = self.get_label_function(sample_pos, sample_scales[scale_ind])

            # Update memory
            self.update_memory(train_x, train_y, learning_rate) ### ?

        # Train filter ### DCF?
        if hard_negative:
            self.filter_optimizer.run(self.params.hard_negative_CG_iter)
        elif (self.frame_num - 1) % self.params.train_skipping == 0:
            self.filter_optimizer.run(self.params.CG_iter)

        timeend = time.time()
        if (self.params.debug >= 3):
            print("Start time:", startitme, " extracted:", time1-startitme, " filtered:", time2-time1, " localized:", time3-time2, " after GIM:", time4-time3, " after update:", timeend-time4, " end:", timeend )
        
        return_state = None
        if self.params.use_segmentation and not self.params.return_bbox and pred_segm_region is not None:
            self.state_tracking = np.array([pred_segm_region[0], pred_segm_region[1], pred_segm_region[2], pred_segm_region[3]])
            return_state = pred_segm_region
        else:
            # Return new state
            new_state = torch.cat((self.pos[[1, 0]] - (self.target_sz[[1, 0]] - 1) / 2, self.target_sz[[1, 0]]))
            self.state_tracking = np.array(new_state.tolist())
            return_state = new_state.tolist()

        # Update RT-MDNet
        # "Long term" update with added ATOM local maxima negatives
        self.conflict = False
        if not self.redetection_mode and self.succesfuly_tracked_frames > 4 and self.frame_num % 5 == 0 and sample_scales[0] > 0.2 and image.shape[0] <= 1080:
            # Find local maxima responses
            print("Sample scales:", sample_scales)
            glob_features = self.extract_processed_image(im, sample_scales, self.img_sample_sz)
            self.global_feat_sz = [torch.tensor(glob_features[0].size()[-2:])] # global feature size, small response matrix representing image
            scores_raw_initial = self.apply_filter(glob_features, self.init_filter)
            scores_raw_full = self.apply_filter(glob_features)
            target_res = torch.tensor([self.im.size()[2], self.im.size()[3]]) # originally self.global_sz
            new_pos_init, s_init = self.localize_target(scores_raw_initial,globl=True,multWindow=False, customResponseRes=target_res)
            new_pos_global, s_global = self.localize_target(scores_raw_full,globl=True,multWindow=False, customResponseRes=target_res)
            response_mtx = s_global[scale_ind, ...].cpu().numpy().squeeze() 
            peak_coords = peak_local_max(response_mtx, min_distance=10, exclude_border=10, num_peaks=30, threshold_rel=0.2)
            # Translate to image coordinates
            peak_coords[:,0] = peak_coords[:,0] / target_res[0].item() * self.im.size()[2] # y 
            peak_coords[:,1] = peak_coords[:,1] / target_res[1].item() * self.im.size()[3] # x

            init_response_mtx = s_init[scale_ind, ...].cpu().numpy().squeeze() 
            init_peak_coords = peak_local_max(init_response_mtx, min_distance=10, exclude_border=10, num_peaks=2, threshold_rel=0.15)
            init_peak_coords[:,0] = init_peak_coords[:,0] / target_res[0].item() * self.im.size()[2] # y 
            init_peak_coords[:,1] = init_peak_coords[:,1] / target_res[1].item() * self.im.size()[3] # x
            
            
            # Build bboxes from peaks.
            bboxes_neg = []
            tsize = np.array([self.state_tracking[3], self.state_tracking[2]]) #height, width
            target_diag = np.linalg.norm(tsize)
            pos_max_npy = self.pos.cpu().numpy()
            added_neg_samps = 10
            # Conflict detection
            visualize = False
            if visualize:
                cv2.imshow("atom_response_init", init_response_mtx)
                cv2.waitKey(5)
            
            # if len(init_peak_coords) == 2:
            #     y0 = int(round(init_peak_coords[0,0]))
            #     x0 = int(round(init_peak_coords[0,1]))
            #     y1 = int(round(init_peak_coords[1,0]))
            #     x1 = int(round(init_peak_coords[1,1]))
            #     y_curr = int(round(pos_max_npy[0]))
            #     x_curr = int(round(pos_max_npy[1]))
            #     if np.linalg.norm(pos_max_npy - init_peak_coords[0,:]) > target_diag/2 and \
            #     init_response_mtx[y0,x0] > 1.25 * init_response_mtx[y1,x1] and \
            #     init_response_mtx[y0,x0] > 1.25 * init_response_mtx[y_curr,x_curr] and \
            #     init_response_mtx[y0,x0] > 0.4:
            #         self.conflict = True
            #         added_neg_samps = 10
            #         proposed_bbox = [init_peak_coords[0,1] - tsize[1]/2, init_peak_coords[0,0] - tsize[0]/2, tsize[1], tsize[0]]
            #         bboxes_neg = generate_samples_neg(image, np.array(self.state_tracking), 30)

            #         self.filter = copy.deepcopy(self.init_filter)
            #         self.pos = torch.tensor([y0,x0])
            #         self.state_tracking = proposed_bbox
            #         return_state = proposed_bbox
            #         self.conflictBox = proposed_bbox
                    
            #         print("CONFLICT FOUND")


            #Hard limit - memory constraint
            if not self.conflict and tsize[0] > 15 and tsize[1] > 15:
                added_neg_samps = 50
                for i in range(0,len(peak_coords)):
                    # filter peaks by distance to target
                    dist_to_max = np.abs(pos_max_npy - peak_coords[i,:])
                    l2_dist = np.linalg.norm(dist_to_max)
                    if dist_to_max[0] < 6*tsize[0] and dist_to_max[1] < 6*tsize[1] and l2_dist > target_diag/2:
                        # Generate 20 negative bboxes around the peak
                        bbox = [peak_coords[i,1] - tsize[1]/2, peak_coords[i,0] - tsize[0]/2, tsize[1], tsize[0]]
                        neg_samps = generate_samples_neg(image, np.array(bbox), 40)
                        bboxes_neg += neg_samps

            # Train the RTMDNet
            print("additional local max negatives:", len(bboxes_neg))
            update_mdnet(self.model, image, np.array(proposed_bbox), np.array([proposed_bbox]), np.array(bboxes_neg) if len(bboxes_neg) > 0 else None, self.mdnet_criterion, self.mdnet_optimizer, 
                self.cropper, self.positive_feats, self.negative_feats, additional_neg_samples=added_neg_samps, additional_pos_samples=49)

        return return_state

    def backtrack(self, visualize=True):
        print("Backtracking...")
        #Initialize DCF on new position
        pars = params.parameters()
        pars.target_not_found_threshold = 0.16
        tracker = Segm(pars)
        tracker.initialize_short_term(self.img_buffer[0], copy.deepcopy(self.state_tracking))

        # Get last frame where true target was still present, see if we still have that frame index in buffer
        lastGoodState = self.state_buffer[0][1]
        lastGoodStateInBuffer = False
        if (self.state_buffer[0][0] > self.frame_num - len(self.img_buffer)):
            lastGoodStateInBuffer = True
        final_ix = len(self.img_buffer)
        if lastGoodStateInBuffer:
            final_ix = self.frame_num - self.state_buffer[0][0] + 1
        
        print("Will backtrack at most " + str(final_ix) + " frames")
        backtrackUntilTargetFound = False
        backtrackUntilFinalFrame = False
        lowConsecutiveScoreCounter = 0
        #For each frame in buffer (or until the frame where last target was seen)
        iter_num = 1
        last_image = None
        backtrack_feat_buf = []
        for i in range(1,final_ix):

            #Track to next state (localize with DCF)
            im = numpy_to_torch(self.img_buffer[i])
            sample_scales = tracker.target_scale * tracker.params.scale_factors # only one scale factor: this is the scale factor between search
            sample_pos = copy.deepcopy(tracker.pos)
            test_x = tracker.extract_processed_sample(im, sample_pos, sample_scales, tracker.img_sample_sz)
            search_region_sz_ = sample_scales[0] * tracker.img_sample_sz
            x0_ = int(round((sample_pos[1] - search_region_sz_[1] / 2).item()))
            y0_ = int(round((sample_pos[0] - search_region_sz_[0] / 2).item()))
            tracker.search_region_ = [x0_, y0_, int(round(search_region_sz_[1].item())), int(round(search_region_sz_[0].item()))]
            scores_raw = tracker.apply_filter(test_x)
            translation_vec, scale_ind, s, flag = tracker.localize_target(scores_raw)
            dcf_score = torch.max(s).item()
            new_pos = sample_pos + translation_vec
            
            #Update position (no scale update, no filter update)
            if new_pos[0] < 0:
                new_pos[0] = 0
            if new_pos[1] < 0:
                new_pos[1] = 0
            if new_pos[0] >= self.img_buffer[i].shape[0]:
                new_pos[0] = self.img_buffer[i].shape[0] - 1
            if new_pos[1] >= self.img_buffer[i].shape[1]:
                new_pos[1] = self.img_buffer[i].shape[1] - 1
            
            #Check if DCF found anything at all
            if flag == "not_found":
                break
    
            tracker.update_state(new_pos)
            # Use GIM as well to get robust new bbox
            segm_region = tracker.segment_target(self.img_buffer[i], new_pos , tracker.target_sz)
            if segm_region is not None:
                tracker.state_tracking = np.array([segm_region[0], segm_region[1], segm_region[2], segm_region[3]])
            else:
                new_state = torch.cat((tracker.pos[[1, 0]] - (tracker.target_sz[[1, 0]] - 1) / 2, tracker.target_sz[[1, 0]]))
                tracker.state_tracking = np.array(new_state.tolist())
            
            #Get MDNet score -- if too low for a few frames in succession, mark as FP
            proposed_bbox = tracker.state_tracking
            score, features = get_score_mdnet(self.img_buffer[i], self.model, np.array(proposed_bbox), self.cropper, proposed_bbox, additional_pos_samps=19)
            backtrack_feat_buf.append(features)
            print(str(i) + ": MDNet:" + str(score))

            #Visualize
            if visualize:
                img = cv2.cvtColor(copy.deepcopy(self.img_buffer[i]), cv2.COLOR_RGB2BGR)
                tl_ = (int(round(proposed_bbox[0])), int(round(proposed_bbox[1])))
                br_ = (int(round(proposed_bbox[0] + proposed_bbox[2])), int(round(proposed_bbox[1] + proposed_bbox[3])))
                pos = (10,25)
                fontColor = (0,255,0) if score > -8 else (0,0,255)
                cv2.putText(img, "MDNet:" + str(round(score*100)/100), 
                    pos, cv2.FONT_HERSHEY_SIMPLEX, 0.7,fontColor,2,2)
                cv2.putText(img, "DCF:" + str(round( dcf_score*100)/100), 
                    (10,95), cv2.FONT_HERSHEY_SIMPLEX, 0.7,(0,255,0),2,2)
                cv2.rectangle(img, tl_, br_, (255, 0, 0), thickness=2)
                cv2.imshow('backtrack', img)
                cv2.waitKey(5)
                last_image = img

            if score < -8:
                lowConsecutiveScoreCounter+=1
            else: 
                lowConsecutiveScoreCounter=0
            if lowConsecutiveScoreCounter >= 5:
                if tracker.state_tracking[0] < 15 or tracker.state_tracking[1] < 15 or \
                    tracker.state_tracking[0]+tracker.state_tracking[2] > self.image_sz[1].item()-15 or \
                    tracker.state_tracking[1]+tracker.state_tracking[3] > self.image_sz[0].item()-15:
                    break
                else:
                    print("Too low MDNet score, returning")
                    if visualize and last_image is not None:
                        pos = (10,60)
                        fontColor = (0,0,255)
                        cv2.putText(last_image, "Too low score. FP.", 
                            pos, cv2.FONT_HERSHEY_SIMPLEX, 0.7,fontColor,2,2)
                        cv2.imshow('backtrack', last_image)
                    return -1, backtrack_feat_buf
            
            iter_num +=1
            if i == final_ix-1:
                if lastGoodStateInBuffer:
                    backtrackUntilTargetFound=True
                else:
                    backtrackUntilFinalFrame=True

        # If made it to end - check if it lines up with position or uncertainty of target at that time
        lastGoodPosition = torch.tensor([lastGoodState[1]+lastGoodState[3]/2, lastGoodState[0]+lastGoodState[2]/2]) #y,x 
        if backtrackUntilTargetFound:
            print("Backtracked. Checking motion model.")
        elif backtrackUntilFinalFrame:
            print("Backtracking limit reached. Checking motion model.")
        else:
            print("Target dissapeared. Checking motion model.")
        prob = self.getProbabilityOfTarget(lastGoodPosition, np.array([lastGoodState[2], lastGoodState[3]]), self.frame_num - iter_num - self.state_buffer[0][0]+1, tracker.pos, visualize=visualize)

        if visualize and last_image is not None:
            pos = (10,60)
            fontColor = (0,255,0) if prob > 0.26 else (0,0,255)
            text = "Lost. Motion prob:"
            if backtrackUntilTargetFound:
                text = "Backtracked. Motion prob:"
            elif backtrackUntilFinalFrame:
                text = "Buffer limit. Motion prob:"
            cv2.putText(last_image, text + str(round(prob*100)/100), 
                pos, cv2.FONT_HERSHEY_SIMPLEX, 0.7,fontColor,2,2)
            cv2.imshow('backtrack', last_image)
        return 1 if prob > 0.26 else 0, backtrack_feat_buf

    def setRedetection(self, state):
        if self.redetection_mode == state:
            return
        else:
            self.redetection_mode = state
            print("Redetection activated" if state else "Redetection deactivated")

    def apply_filter(self, sample_x: TensorList, custom_filter=None):
        if custom_filter is None:
            return operation.conv2d(sample_x, self.filter, mode='same')
        else:
            return operation.conv2d(sample_x, custom_filter, mode='same')

    def psr(self, response):
        max_val, max_ind = dcf.max2d(response)
        row = max_ind[0].cpu().view(-1)[0].item()
        col = max_ind[0].cpu().view(-1)[1].item()
        x1 = np.minimum(response.size()[-1], np.maximum(1, col - np.round(0.05 * response.size()[-2])))
        x2 = np.minimum(response.size()[-1], np.maximum(1, col + np.round(0.05 * response.size()[-2])))
        y1 = np.minimum(response.size()[-1], np.maximum(1, row - np.round(0.05 * response.size()[-1])))
        y2 = np.minimum(response.size()[-1], np.maximum(1, row + np.round(0.05 * response.size()[-1])))

        resp_npy = response.cpu().numpy().squeeze()
        M = np.ones(resp_npy.shape)
        M[int(y1):int(y2), int(x1):int(x2)] = 0
        sidelobe = resp_npy[M==1]
        mu_s = np.mean(sidelobe)
        sigma_s = np.std(sidelobe)
        s = (max_val.item() - mu_s) / (sigma_s + 0.0001)
        s = s * max_val.item()
        return s

    def localize_target(self, scores_raw, globl=False, multWindow=True, customResponseRes=None):
        # Weighted sum (if multiple features) with interpolation in fourier domain
        weight = self.fparams.attribute('translation_weight', 1.0)
        scores_raw = weight * scores_raw
        sf_weighted = fourier.cfft2(scores_raw) / (scores_raw.size(2) * scores_raw.size(3)) #[1, 1, 15, 8, 2] 
        feat_sz = self.feature_sz if not globl else self.global_feat_sz
        for i, (sz, ksz) in enumerate(zip(feat_sz, self.kernel_size)):
            sf_weighted[i] = fourier.shift_fs(sf_weighted[i],
                                              math.pi * (1 - torch.Tensor([ksz[0] % 2, ksz[1] % 2]) / sz))

        scores_fs = fourier.sum_fs(sf_weighted) #[1, 1, 15, 8, 2]
        resolution = customResponseRes
        if resolution == None:
            resolution = self.output_sz if not globl else self.global_sz
        scores = fourier.sample_fs(scores_fs, resolution) # 1 x 1 x 224 x 224

        if self.output_window is not None and not getattr(self.params, 'perform_hn_without_windowing', False) and not globl:
            scores *= self.output_window

        if getattr(self.params, 'advanced_localization', False) and not globl:
            return self.localize_advanced(scores) ### ?

        if globl:
            return self.localize_global(scores, multWindow)

        # Get maximum
        max_score, max_disp = dcf.max2d(scores)
        _, scale_ind = torch.max(max_score, dim=0)
        max_disp = max_disp.float().cpu()

        # Convert to displacements in the base scale
        disp = (max_disp + self.output_sz / 2) % self.output_sz - self.output_sz / 2

        # Compute translation vector and scale change factor
        translation_vec = disp[scale_ind, ...].view(-1) * (self.img_support_sz / self.output_sz) * self.target_scale
        translation_vec *= self.params.scale_factors[scale_ind]

        # Shift the score output for visualization purposes
        if self.params.debug >= 2:
            sz = scores.shape[-2:]
            scores = torch.cat([scores[..., sz[0] // 2:, :], scores[..., :sz[0] // 2, :]], -2)
            scores = torch.cat([scores[..., :, sz[1] // 2:], scores[..., :, :sz[1] // 2]], -1)

        return translation_vec, scale_ind, scores, None

    def localize_advanced(self, scores):
        """Does the advanced localization with hard negative detection and target not found."""

        sz = scores.shape[-2:]

        if self.output_window is not None and getattr(self.params, 'perform_hn_without_windowing', False):
            scores_orig = scores.clone()

            scores_orig = torch.cat([scores_orig[..., (sz[0] + 1) // 2:, :],
                                     scores_orig[..., :(sz[0] + 1) // 2, :]], -2)
            scores_orig = torch.cat([scores_orig[..., :, (sz[1] + 1) // 2:],
                                     scores_orig[..., :, :(sz[1] + 1) // 2]], -1)

            scores *= self.output_window

        # Shift scores back
        scores = torch.cat([scores[..., (sz[0] + 1) // 2:, :], scores[..., :(sz[0] + 1) // 2, :]], -2)
        scores = torch.cat([scores[..., :, (sz[1] + 1) // 2:], scores[..., :, :(sz[1] + 1) // 2]], -1)

        # Find maximum
        max_score1, max_disp1 = dcf.max2d(scores)
        _, scale_ind = torch.max(max_score1, dim=0)
        max_score1 = max_score1[scale_ind]
        max_disp1 = max_disp1[scale_ind, ...].float().cpu().view(-1)
        target_disp1 = max_disp1 - self.output_sz // 2
        translation_vec1 = target_disp1 * (self.img_support_sz / self.output_sz) * self.target_scale

        self.max_val = max_score1.item()

        if max_score1.item() < self.params.target_not_found_threshold:
            print('Target is not found')
            return translation_vec1, scale_ind, scores, 'not_found'

        if self.output_window is not None and getattr(self.params, 'perform_hn_without_windowing', False):
            scores = scores_orig

        # Mask out target neighborhood
        target_neigh_sz = self.params.target_neighborhood_scale * self.target_sz / self.target_scale
        tneigh_top = max(round(max_disp1[0].item() - target_neigh_sz[0].item() / 2), 0)
        tneigh_bottom = min(round(max_disp1[0].item() + target_neigh_sz[0].item() / 2 + 1), sz[0])
        tneigh_left = max(round(max_disp1[1].item() - target_neigh_sz[1].item() / 2), 0)
        tneigh_right = min(round(max_disp1[1].item() + target_neigh_sz[1].item() / 2 + 1), sz[1])
        scores_masked = scores[scale_ind:scale_ind + 1, ...].clone()
        scores_masked[..., tneigh_top:tneigh_bottom, tneigh_left:tneigh_right] = 0

        # Find new maximum
        max_score2, max_disp2 = dcf.max2d(scores_masked)
        max_disp2 = max_disp2.float().cpu().view(-1)
        target_disp2 = max_disp2 - self.output_sz // 2
        translation_vec2 = target_disp2 * (self.img_support_sz / self.output_sz) * self.target_scale

        self.max_1 = max_score1.item()
        self.max_2 = max_score2.item()
        self.translation_1 = translation_vec1
        self.translation_2 = translation_vec2

        # Handle the different cases
        if max_score2 > self.params.distractor_threshold * max_score1:
            disp_norm1 = torch.sqrt(torch.sum(target_disp1 ** 2))
            disp_norm2 = torch.sqrt(torch.sum(target_disp2 ** 2))
            disp_threshold = self.params.dispalcement_scale * math.sqrt(sz[0] * sz[1]) / 2

            if disp_norm2 > disp_threshold and disp_norm1 < disp_threshold:
                return translation_vec1, scale_ind, scores, 'hard_negative'
            if disp_norm2 < disp_threshold and disp_norm1 > disp_threshold:
                return translation_vec2, scale_ind, scores, 'hard_negative'
            if disp_norm2 > disp_threshold and disp_norm1 > disp_threshold:
                return translation_vec1, scale_ind, scores, 'uncertain'

            # If also the distractor is close, return with highest score
            return translation_vec1, scale_ind, scores, 'uncertain'

        if max_score2 > self.params.hard_negative_threshold * max_score1 and \
                        max_score2 > self.params.target_not_found_threshold:
            # print('%d.) Hard negative detected. max1=%.3f, max2=%.3f, ratio=%.2f' %
            #       (self.frame_num, max_score1, max_score2, max_score2 / max_score1))
            if (max_score2 / max_score1) < self.params.peaks_ratio:
                return translation_vec1, scale_ind, scores, 'hard_negative'
            else:
                return translation_vec1, scale_ind, scores, 'uncertain'

        return translation_vec1, scale_ind, scores, None
    
    def getProbabilityOfTarget(self, good_position, good_size, frame_difference, new_position, visualize=True):
        size_factor = 1.05 ** frame_difference
        sigma_factor = 0.5
        y_pixel = np.round(good_position[0].item())
        x_pixel = np.round(good_position[1].item())
        y_,x_ = torch.meshgrid(torch.arange(0,self.image_sz[0].item())-y_pixel, torch.arange(0,self.image_sz[1].item())-x_pixel)
        obj_size = good_size #w,h
        g = np.exp(-0.5 * (x_**2 / ((sigma_factor*size_factor*obj_size[0])**2) + y_**2 / ((sigma_factor*size_factor*obj_size[1])**2) ) )
        y_new = int(np.round(new_position[0].item()))
        x_new = int(np.round(new_position[1].item()))
        if y_new < 0:
            y_new = 0
        if x_new < 0:
            x_new = 0
        if y_new >= self.image_sz[0].item():
            y_new = int(self.image_sz[0].item() - 1)
        if x_new >= self.image_sz[1].item():
            x_new = int(self.image_sz[1].item() - 1)
        if visualize:
            cv2.imshow("uncert",g.cpu().numpy())
        return g[y_new,x_new].item()

    def getGaussianWindow(self, size):
        size_factor = 1.05 ** (self.undetected_frames-1)
        sigma_factor = 0.5
        y_pixel = np.round(self.last_good_position[0].item() / self.image_sz[0].item() * size[0])
        x_pixel = np.round(self.last_good_position[1].item() / self.image_sz[1].item() * size[1])
        y_,x_ = torch.meshgrid(torch.arange(0,size[0])-y_pixel, torch.arange(0,size[1])-x_pixel)
        obj_size = self.base_target_sz.cpu().numpy() #h,w
        g = torch.exp(-0.5 * (x_**2 / ((sigma_factor*size_factor*obj_size[1])**2) + y_**2 / ((sigma_factor*size_factor*obj_size[0])**2) ) )
        g = g.to(self.params.device)
        return g

    def localize_global(self, scores, multWindow=True):
        sz = scores.shape[-2:]
        
        # Shift scores back # CENTERING
        scores = torch.cat([scores[..., (sz[0] + 1) // 2:, :], scores[..., :(sz[0] + 1) // 2, :]], -2)
        scores = torch.cat([scores[..., :, (sz[1] + 1) // 2:], scores[..., :, :(sz[1] + 1) // 2]], -1)

        window = self.getGaussianWindow(sz)
        if multWindow:
            scores = scores * window
        # Find maximum
        max_score1, max_disp1 = dcf.max2d(scores)
        _, scale_ind = torch.max(max_score1, dim=0)
        max_score1 = max_score1[scale_ind]
        max_disp1 = max_disp1[scale_ind, ...].float().cpu().view(-1)

        max_position_orig = [0,0]
        max_position_orig[0] = max_disp1[0].item() / self.global_sz[0].item() * self.im.size()[2] # transform back into image coordinates
        max_position_orig[1] = max_disp1[1].item() / self.global_sz[1].item() * self.im.size()[3]

        return torch.tensor(max_position_orig), scores

    def extract_sample(self, im: torch.Tensor, pos: torch.Tensor, scales, sz: torch.Tensor):
        return self.params.features_filter.extract(im, pos, scales, sz)

    def extract_processed_sample(self, im: torch.Tensor, pos: torch.Tensor, scales, sz: torch.Tensor) -> (
    TensorList, TensorList):
        x = self.extract_sample(im, pos, scales, sz)
        return self.preprocess_sample(self.project_sample(x))

    def extract_processed_image(self, im: torch.Tensor, scales, sz: torch.Tensor):
        x, imsize = self.params.features_filter.extract_full_img(im, scales, sz)
        self.global_sz = torch.tensor(imsize[-2:])
        return self.project_sample(x)

    def preprocess_sample(self, x: TensorList) -> (TensorList, TensorList):
        if getattr(self.params, '_feature_window', False):
            x = x * self.feature_window
        return x

    def project_sample(self, x: TensorList, proj_matrix=None):
        # Apply projection matrix
        if proj_matrix is None:
            proj_matrix = self.projection_matrix
        return operation.conv2d(x, proj_matrix).apply(self.projection_activation)

    def init_learning(self):
        # Get window function
        self.feature_window = TensorList([dcf.hann2d(sz).to(self.params.device) for sz in self.feature_sz])

        # Filter regularization
        self.filter_reg = self.fparams.attribute('filter_reg')

        # Activation function after the projection matrix (phi_1 in the paper)
        projection_activation = getattr(self.params, 'projection_activation', 'none')
        if isinstance(projection_activation, tuple):
            projection_activation, act_param = projection_activation

        if projection_activation == 'none':
            self.projection_activation = lambda x: x
        elif projection_activation == 'relu':
            self.projection_activation = torch.nn.ReLU(inplace=True)
        elif projection_activation == 'elu':
            self.projection_activation = torch.nn.ELU(inplace=True)
        elif projection_activation == 'mlu':
            self.projection_activation = lambda x: F.elu(F.leaky_relu(x, 1 / act_param), act_param)
        else:
            raise ValueError('Unknown activation')

        # Activation function after the output scores (phi_2 in the paper)
        response_activation = getattr(self.params, 'response_activation', 'none')
        if isinstance(response_activation, tuple):
            response_activation, act_param = response_activation

        if response_activation == 'none':
            self.response_activation = lambda x: x
        elif response_activation == 'relu':
            self.response_activation = torch.nn.ReLU(inplace=True)
        elif response_activation == 'elu':
            self.response_activation = torch.nn.ELU(inplace=True)
        elif response_activation == 'mlu':
            self.response_activation = lambda x: F.elu(F.leaky_relu(x, 1 / act_param), act_param)
        else:
            raise ValueError('Unknown activation')

    def generate_init_samples(self, im: torch.Tensor) -> TensorList:
        """Generate augmented initial samples."""

        # Compute augmentation size
        aug_expansion_factor = getattr(self.params, 'augmentation_expansion_factor', None)
        aug_expansion_sz = self.img_sample_sz.clone()
        aug_output_sz = None
        if aug_expansion_factor is not None and aug_expansion_factor != 1:
            aug_expansion_sz = (self.img_sample_sz * aug_expansion_factor).long()
            aug_expansion_sz += (aug_expansion_sz - self.img_sample_sz.long()) % 2
            aug_expansion_sz = aug_expansion_sz.float()
            aug_output_sz = self.img_sample_sz.long().tolist()

        # Random shift operator
        get_rand_shift = lambda: None
        random_shift_factor = getattr(self.params, 'random_shift_factor', 0)
        if random_shift_factor > 0:
            get_rand_shift = lambda: ((torch.rand(2) - 0.5) * self.img_sample_sz * random_shift_factor).long().tolist()

        # Create transofmations
        self.transforms = [augmentation.Identity(aug_output_sz)]
        if 'shift' in self.params.augmentation:
            self.transforms.extend(
                [augmentation.Translation(shift, aug_output_sz) for shift in self.params.augmentation['shift']])
        if 'relativeshift' in self.params.augmentation:
            get_absolute = lambda shift: (torch.Tensor(shift) * self.img_sample_sz / 2).long().tolist()
            self.transforms.extend([augmentation.Translation(get_absolute(shift), aug_output_sz) for shift in
                                    self.params.augmentation['relativeshift']])
        if 'fliplr' in self.params.augmentation and self.params.augmentation['fliplr']:
            self.transforms.append(augmentation.FlipHorizontal(aug_output_sz, get_rand_shift()))
        if 'blur' in self.params.augmentation:
            self.transforms.extend([augmentation.Blur(sigma, aug_output_sz, get_rand_shift()) for sigma in
                                    self.params.augmentation['blur']])
        if 'scale' in self.params.augmentation:
            self.transforms.extend(
                [augmentation.Scale(scale_factor, aug_output_sz, get_rand_shift()) for scale_factor in
                 self.params.augmentation['scale']])
        if 'rotate' in self.params.augmentation:
            self.transforms.extend([augmentation.Rotate(angle, aug_output_sz, get_rand_shift()) for angle in
                                    self.params.augmentation['rotate']])

        init_samples = self.params.features_filter.extract_transformed(im, self.pos.round(), self.target_scale,
                                                                       aug_expansion_sz, self.transforms)

        # Remove augmented samples for those that shall not have
        for i, use_aug in enumerate(self.fparams.attribute('use_augmentation')):
            if not use_aug:
                init_samples[i] = init_samples[i][0:1, ...]

        # Add dropout samples
        if 'dropout' in self.params.augmentation:
            num, prob = self.params.augmentation['dropout']
            self.transforms.extend(self.transforms[:1] * num)
            for i, use_aug in enumerate(self.fparams.attribute('use_augmentation')):
                if use_aug:
                    init_samples[i] = torch.cat([init_samples[i],
                                                 F.dropout2d(init_samples[i][0:1, ...].expand(num, -1, -1, -1), p=prob,
                                                             training=True)])

        return init_samples

    def init_projection_matrix(self, x):
        # Set if using projection matrix
        self.params.use_projection_matrix = getattr(self.params, 'use_projection_matrix', True)

        if self.params.use_projection_matrix:
            self.compressed_dim = self.fparams.attribute('compressed_dim', None)

            proj_init_method = getattr(self.params, 'proj_init_method', 'pca')
            if proj_init_method == 'pca':
                x_mat = TensorList([e.permute(1, 0, 2, 3).reshape(e.shape[1], -1).clone() for e in x])
                x_mat -= x_mat.mean(dim=1, keepdim=True)
                cov_x = x_mat @ x_mat.t()
                self.projection_matrix = TensorList(
                    [None if cdim is None else torch.svd(C)[0][:, :cdim].t().unsqueeze(-1).unsqueeze(-1).clone() for
                     C, cdim in
                     zip(cov_x, self.compressed_dim)])
            elif proj_init_method == 'randn':
                self.projection_matrix = TensorList(
                    [None if cdim is None else ex.new_zeros(cdim, ex.shape[1], 1, 1).normal_(0,
                                                                                             1 / math.sqrt(ex.shape[1]))
                     for ex, cdim in
                     zip(x, self.compressed_dim)])
        else:
            self.compressed_dim = x.size(1)
            self.projection_matrix = TensorList([None] * len(x))

    def init_label_function(self, train_x):
        # Allocate label function
        self.y = TensorList([x.new_zeros(self.params.sample_memory_size, 1, x.shape[2], x.shape[3]) for x in train_x])

        # Output sigma factor
        output_sigma_factor = self.fparams.attribute('output_sigma_factor')
        self.sigma = (
                     self.feature_sz / self.img_support_sz * self.base_target_sz).prod().sqrt() * output_sigma_factor * torch.ones(
            2)

        # Center pos in normalized coords
        target_center_norm = (self.pos - self.pos.round()) / (self.target_scale * self.img_support_sz)

        # Generate label functions
        for y, sig, sz, ksz, x in zip(self.y, self.sigma, self.feature_sz, self.kernel_size, train_x):
            center_pos = sz * target_center_norm + 0.5 * torch.Tensor([(ksz[0] + 1) % 2, (ksz[1] + 1) % 2])
            for i, T in enumerate(self.transforms[:x.shape[0]]):
                sample_center = center_pos + torch.Tensor(T.shift) / self.img_support_sz * sz
                y[i, 0, ...] = dcf.label_function_spatial(sz, sig, sample_center)

        # Return only the ones to use for initial training
        return TensorList([y[:x.shape[0], ...] for y, x in zip(self.y, train_x)])

    def init_memory(self, train_x):
        # Initialize first-frame training samples
        self.num_init_samples = train_x.size(0)
        self.init_sample_weights = TensorList([x.new_ones(1) / x.shape[0] for x in train_x])
        self.init_training_samples = train_x

        # Sample counters and weights
        self.num_stored_samples = self.num_init_samples.copy()
        self.previous_replace_ind = [None] * len(self.num_stored_samples)
        self.sample_weights = TensorList([x.new_zeros(self.params.sample_memory_size) for x in train_x])
        for sw, init_sw, num in zip(self.sample_weights, self.init_sample_weights, self.num_init_samples):
            sw[:num] = init_sw

        # Initialize memory
        self.training_samples = TensorList(
            [x.new_zeros(self.params.sample_memory_size, cdim, x.shape[2], x.shape[3]) for x, cdim in
             zip(train_x, self.compressed_dim)])

    def update_memory(self, sample_x: TensorList, sample_y: TensorList, learning_rate=None):
        replace_ind = self.update_sample_weights(self.sample_weights, self.previous_replace_ind,
                                                 self.num_stored_samples, self.num_init_samples, self.fparams,
                                                 learning_rate)
        self.previous_replace_ind = replace_ind
        for train_samp, x, ind in zip(self.training_samples, sample_x, replace_ind):
            train_samp[ind:ind + 1, ...] = x
        for y_memory, y, ind in zip(self.y, sample_y, replace_ind):
            y_memory[ind:ind + 1, ...] = y
        if self.hinge_mask is not None:
            for m, y, ind in zip(self.hinge_mask, sample_y, replace_ind):
                m[ind:ind + 1, ...] = (y >= self.params.hinge_threshold).float()
        self.num_stored_samples += 1

    def update_sample_weights(self, sample_weights, previous_replace_ind, num_stored_samples, num_init_samples, fparams,
                              learning_rate=None):
        # Update weights and get index to replace in memory
        replace_ind = []
        for sw, prev_ind, num_samp, num_init, fpar in zip(sample_weights, previous_replace_ind, num_stored_samples,
                                                          num_init_samples, fparams):
            lr = learning_rate
            if lr is None:
                lr = fpar.learning_rate

            init_samp_weight = getattr(fpar, 'init_samples_minimum_weight', None)
            if init_samp_weight == 0:
                init_samp_weight = None
            s_ind = 0 if init_samp_weight is None else num_init

            if num_samp == 0 or lr == 1:
                sw[:] = 0
                sw[0] = 1
                r_ind = 0
            else:
                # Get index to replace
                _, r_ind = torch.min(sw[s_ind:], 0)
                r_ind = r_ind.item() + s_ind

                # Update weights
                if prev_ind is None:
                    sw /= 1 - lr
                    sw[r_ind] = lr
                else:
                    sw[r_ind] = sw[prev_ind] / (1 - lr)

            sw /= sw.sum()
            if init_samp_weight is not None and sw[:num_init].sum() < init_samp_weight:
                sw /= init_samp_weight + sw[num_init:].sum()
                sw[:num_init] = init_samp_weight / num_init

            replace_ind.append(r_ind)

        return replace_ind

    def get_label_function(self, sample_pos, sample_scale):
        # Generate label function
        train_y = TensorList()
        target_center_norm = (self.pos - sample_pos) / (sample_scale * self.img_support_sz)
        for sig, sz, ksz in zip(self.sigma, self.feature_sz, self.kernel_size):
            center = sz * target_center_norm + 0.5 * torch.Tensor([(ksz[0] + 1) % 2, (ksz[1] + 1) % 2])
            train_y.append(dcf.label_function_spatial(sz, sig, center))
        return train_y

    def update_state(self, new_pos, new_scale=None):
        # Update scale
        if new_scale is not None:
            self.target_scale = new_scale.clamp(self.min_scale_factor, self.max_scale_factor)
            self.target_sz = self.base_target_sz * self.target_scale

        # Update pos
        inside_ratio = 0.2
        inside_offset = (inside_ratio - 0.5) * self.target_sz
        self.pos = torch.max(torch.min(new_pos, self.image_sz - inside_offset), inside_offset)

    def create_dist(self, width, height, cx=None, cy=None):

        if cx is None:
            cx = width / 2
        if cy is None:
            cy = height / 2

        x_ = np.linspace(1, width, width) - cx
        y_ = np.linspace(1, width, width) - cy
        X, Y = np.meshgrid(x_, y_)

        return np.sqrt(np.square(X) + np.square(Y)).astype(np.float32)

    def init_segmentation(self, image, bb, init_mask=None):
        # crop image patch
        init_patch_crop, f_ = prutils.sample_target(image, np.array(bb), self.params.segm_search_area_factor,
                                                    output_sz=self.params.segm_output_sz)

        self.segmentation_task = False
        if init_mask is not None:
            mask = copy.deepcopy(init_mask).astype(np.float32)

            # uncomment this if evaluating on video object segmentation task (e.g., DAVIS)
            #self.segmentation_task = True
            # self.params.segm_optimize_polygon = True
            # segmentation videos are shorter - therefore larger scale change factor can be used
            #self.params.min_scale_change_factor = 0.9
            #self.params.max_scale_change_factor = 1.1
            #self.params.segm_mask_thr = 0.2

            # used for tracking-by-segmentation (use commented lines above for VOS)
            self.segmentation_task = False
        else:
            mask = np.zeros((image.shape[0], image.shape[1]), dtype=np.int32)
            if hasattr(self, 'gt_poly'):
                p1 = self.gt_poly[:2]
                p2 = self.gt_poly[2:4]
                p3 = self.gt_poly[4:6]
                p4 = self.gt_poly[6:]
                cv2.fillConvexPoly(mask, np.array([p1, p2, p3, p4], dtype=np.int32), 1)
                mask = mask.astype(np.float32)
            else:
                p1 = bb[:2]
                p2 = [bb[0] + bb[2], bb[1]]
                p3 = [bb[0] + bb[2], bb[1] + bb[3]]
                p4 = [bb[0], bb[1] + bb[3]]
                cv2.fillConvexPoly(mask, np.array([p1, p2, p3, p4], dtype=np.int32), 1)
                mask = mask.astype(np.float32)

        init_mask_patch_np, patch_factor_init = prutils.sample_target(mask, np.array(bb),
                                                                      self.params.segm_search_area_factor,
                                                                      output_sz=self.params.segm_output_sz, pad_val=0)

        # network was renamed therefore we need to specify constructor_module and constructor_fun_name
        segm_net, _ = load_network(self.params.bbox_reg_net_path, backbone_pretrained=self.params.resnet_octconv_path,
                                   constructor_module='ltr.models.bbox_reg.bbox_reg',
                                   constructor_fun_name='bbox_reg_resnet50')

        if self.params.use_gpu:
            segm_net.cuda()
        segm_net.eval()

        for p in segm_net.segm_predictor.parameters():
            p.requires_grad = False

        for p in segm_net.bbox_predictor.parameters():
            p.requires_grad = False

        self.params.segm_normalize_mean = np.array(self.params.segm_normalize_mean).reshape((1, 1, 3))
        self.params.segm_normalize_std = np.array(self.params.segm_normalize_std).reshape((1, 1, 3))

        # normalize input image
        init_patch_norm_ = init_patch_crop.astype(np.float32) / float(255)
        init_patch_norm_ -= self.params.segm_normalize_mean
        init_patch_norm_ /= self.params.segm_normalize_std

        # create distance map for discriminative segmentation
        if self.params.segm_use_dist:
            if self.params.segm_dist_map_type == 'center':
                # center-based dist map
                dist_map = self.create_dist(init_patch_crop.shape[0], init_patch_crop.shape[1])
            else:
                print('Error: Unknown distance map type.')
                exit(-1)

            dist_map = torch.Tensor(dist_map)

        # put image patch and mask to GPU
        init_patch = torch.Tensor(init_patch_norm_)
        init_mask_patch = torch.Tensor(init_mask_patch_np)
        if self.params.use_gpu:
            init_patch = init_patch.to(self.params.device)
            init_mask_patch = init_mask_patch.to(self.params.device)
            if self.params.segm_use_dist:
                dist_map = dist_map.to(self.params.device)
                dist_map = torch.unsqueeze(torch.unsqueeze(dist_map, dim=0), dim=0)
                test_dist_map = [dist_map]
            else:
                test_dist_map = None

        # reshape image for the feature extractor
        init_patch = torch.unsqueeze(init_patch, dim=0).permute(0, 3, 1, 2)
        init_mask_patch = torch.unsqueeze(torch.unsqueeze(init_mask_patch, dim=0), dim=0)

        # extract features (extracting twice on the same patch - not necessary)
        train_feat = segm_net.extract_backbone_features(init_patch)

        # prepare features in the list (format for the network)
        train_feat_segm = [feat for feat in train_feat.values()]
        test_feat_segm = [feat for feat in train_feat.values()]
        train_masks = [init_mask_patch]

        if init_mask is None:
            iters = 0
            while iters < 1:
                # Obtain segmentation prediction
                segm_pred, _ = segm_net.segm_predictor(test_feat_segm, train_feat_segm, train_masks, test_dist_map)

                # softmax on the prediction (during training this is done internaly when calculating loss)
                # take only the positive channel as predicted segmentation mask
                mask_real = F.softmax(segm_pred, dim=1)[0, 0, :, :].cpu().numpy()
                mask = (mask_real > self.params.init_segm_mask_thr).astype(np.float32)

                if hasattr(self, 'gt_poly'):
                    # dilate polygon-based mask
                    # dilate only if given mask is made from polygon, not from axis-aligned bb (since rotated bb is much tighter)
                    dil_kernel_sz = max(5, int(round(0.05 * min(self.target_sz).item() * f_)))
                    kernel = np.ones((dil_kernel_sz, dil_kernel_sz), np.uint8)
                    mask_dil = cv2.dilate(init_mask_patch_np, kernel, iterations=1)
                    mask = mask * mask_dil
                else:
                    mask = mask * init_mask_patch_np

                target_pixels = np.sum((mask > 0.5).astype(np.float32))
                self.segm_init_target_pixels = target_pixels

                if self.params.save_mask:
                    segm_crop_sz = math.ceil(math.sqrt(bb[2] * bb[3]) * self.params.segm_search_area_factor)
                    save_mask(None, mask_real, segm_crop_sz, bb, image.shape[1], image.shape[0],
                              self.params.masks_save_path, self.sequence_name, self.frame_name)
                else:
                    segm_crop_sz = math.ceil(math.sqrt(bb[2] * bb[3]) * self.params.segm_search_area_factor)
                    self.pred_mask = save_mask(None, mask_real, segm_crop_sz, bb, image.shape[1], image.shape[0],
                              None, None, None)

                mask_gpu = torch.unsqueeze(torch.unsqueeze(torch.tensor(mask), dim=0), dim=0).to(self.params.device)
                train_masks = [mask_gpu]

                iters += 1
        else:
            init_mask_patch_np = (init_mask_patch_np > 0.1).astype(np.float32)
            target_pixels = np.sum((init_mask_patch_np).astype(np.float32))
            self.segm_init_target_pixels = target_pixels

            mask_gpu = torch.unsqueeze(torch.unsqueeze(torch.tensor(init_mask_patch_np), dim=0), dim=0).to(
                self.params.device)

            self.pred_mask = init_mask

        if self.params.bbox_reg_level == 3:
            points_ = Point(16, 16, self.params.segm_output_sz // 2)
        elif self.params.bbox_reg_level == 2:
            points_ = Point(8, 36, self.params.segm_output_sz // 2)
        else:
            print('Error: bbox regression level unknown.')
            exit(-1)
        self.points_x_vec = np.reshape(points_.points[0], (-1))
        self.points_y_vec = np.reshape(points_.points[1], (-1))

        # store everything that is needed for later
        self.segm_net = segm_net
        self.train_feat_segm = train_feat_segm
        self.init_mask_patch = mask_gpu
        if self.params.segm_use_dist:
            self.dist_map = dist_map

    def segment_target(self, image, pos, sz):
        # pos and sz are in the image coordinates
        # construct new bounding box first
        tlx_ = pos[1] - sz[1] / 2
        tly_ = pos[0] - sz[0] / 2
        w_ = sz[1]
        h_ = sz[0]
        bb = [tlx_.item(), tly_.item(), w_.item(), h_.item()]

        patch, f_ = prutils.sample_target(image, np.array(bb), self.params.segm_search_area_factor,
                                          output_sz=self.params.segm_output_sz)

        segm_crop_sz = math.ceil(math.sqrt(bb[2] * bb[3]) * self.params.segm_search_area_factor)

        # normalize input image
        init_patch_norm_ = patch.astype(np.float32) / float(255)
        init_patch_norm_ -= self.params.segm_normalize_mean
        init_patch_norm_ /= self.params.segm_normalize_std

        # put image patch and mask to GPU
        patch_gpu = torch.Tensor(init_patch_norm_)
        if self.params.use_gpu:
            patch_gpu = patch_gpu.to(self.params.device)

            # reshape image for the feature extractor
            patch_gpu = torch.unsqueeze(patch_gpu, dim=0).permute(0, 3, 1, 2)

        # extract features (extracting twice on the same patch - not necessary)
        test_feat = self.segm_net.extract_backbone_features(patch_gpu)

        # prepare features in the list (format for the network)
        test_feat_segm = [feat for feat in test_feat.values()]
        train_masks = [self.init_mask_patch]
        if self.params.segm_use_dist:
            if self.params.segm_dist_map_type == 'center':
                # center-based distance map
                test_dist_map = [self.dist_map]
            elif self.params.segm_dist_map_type == 'bbox':
                # bbox-based distance map
                D = self.create_dist_gauss(self.params.segm_output_sz, w_.item() * f_, h_.item() * f_)
                test_dist_map = [torch.unsqueeze(torch.unsqueeze(torch.Tensor(D).to(self.params.device), dim=0), dim=0)]
        else:
            test_dist_map = None

        # Obtain segmentation prediction
        segm_pred, _ = self.segm_net.segm_predictor(test_feat_segm, self.train_feat_segm, train_masks, test_dist_map)

        # get bounding box prediction
        if self.params.bbox_reg_for_scale:
            mask_for_bbox = torch.unsqueeze(F.softmax(segm_pred, dim=1)[:, 0, :, :], dim=1)

            bbox_reg, bbox_cls = self.segm_net.bbox_predictor(self.train_feat_segm, test_feat_segm, mask_for_bbox)

            cls_sm = bbox_cls[0].softmax(-1)#.detach().cpu().numpy().astype(np.float32)

            reg_vec = bbox_reg[0].view(4, -1).cpu().detach().numpy().astype(np.float32)
            cls_sm_vec = cls_sm[1].view(-1).detach().cpu().numpy().astype(np.float32)

            reg_x0 = self.points_x_vec - reg_vec[0, :]
            reg_y0 = self.points_y_vec - reg_vec[1, :]
            reg_x1 = self.points_x_vec + reg_vec[2, :]
            reg_y1 = self.points_y_vec + reg_vec[3, :]

            reg_cx = (reg_x0 + reg_x1) / 2
            reg_cy = (reg_y0 + reg_y1) / 2
            reg_w = reg_x1 - reg_x0
            reg_h = reg_y1 - reg_y0

            penalty_k = 0.14
            win_influence = 0.45
            if self.params.tracking:
               track_lr = 0.30
            else:
               track_lr = 1

            # scale penalty
            s_c = calc_change(calc_sz(reg_w, reg_h) / (calc_sz(self.bb_sz[0] * f_, self.bb_sz[1] * f_)))

            # aspect ratio penalty
            r_c = calc_change((self.bb_sz[0] / self.bb_sz[1]) / (reg_w / reg_h))
            penalty = np.exp(-(r_c * s_c - 1) * penalty_k)
            pscore = penalty * cls_sm_vec

            # window penalty
            pscore = pscore * (1 - win_influence) + self.bb_reg_window * win_influence
            best_idx = np.argmax(pscore)
            center_w_h = [reg_cx[best_idx] / f_, reg_cy[best_idx] / f_, reg_w[best_idx] / f_, reg_h[best_idx] / f_]

            lr = penalty[best_idx] * cls_sm_vec[best_idx] * track_lr

            cx_ = (center_w_h[0] - 192 / f_) + pos[1].item()
            cy_ = (center_w_h[1] - 192 / f_) + pos[0].item()

            # smooth bbox
            width_ = self.bb_sz[0] * (1 - lr) + center_w_h[2] * lr
            height_ = self.bb_sz[1] * (1 - lr) + center_w_h[3] * lr

            self.target_scale *= np.sqrt(width_ * height_) / np.sqrt(self.bb_sz[0] * self.bb_sz[1])  # uncomment
            self.target_scale = self.target_scale.clamp(self.min_scale_factor, self.max_scale_factor)

            self.bb_sz = np.array([width_, height_])
            self.reg_bb = [cx_ - width_ / 2, cy_ - height_ / 2, width_, height_]

            self.target_sz = self.base_target_sz * self.target_scale  # uncomment

        # softmax on the prediction (during training this is done internaly when calculating loss)
        # take only the positive channel as predicted segmentation mask
        mask = F.softmax(segm_pred, dim=1)[0, 0, :, :].cpu().numpy()

        mask_real = copy.copy(mask)
        mask = (mask > self.params.segm_mask_thr).astype(np.uint8)

        if cv2.__version__[-5] == '4':
            contours, _ = cv2.findContours(mask, cv2.RETR_EXTERNAL, cv2.CHAIN_APPROX_NONE)
        else:
            _, contours, _ = cv2.findContours(mask, cv2.RETR_EXTERNAL, cv2.CHAIN_APPROX_NONE)
        cnt_area = [cv2.contourArea(cnt) for cnt in contours]

        if self.segmentation_task:
            mask = np.zeros(mask.shape, dtype=np.uint8)
            cv2.drawContours(mask, contours, -1, 1, thickness=-1)

            # save mask to disk
            # Note: move this below if evaluating on VOT
            if self.params.save_mask:
                save_mask(None, mask_real, segm_crop_sz, bb, image.shape[1], image.shape[0],
                          self.params.masks_save_path, self.sequence_name, self.frame_name,
                          segm_threshold=self.params.segm_mask_thr)
            else:
                self.pred_mask = save_mask(None, mask_real, segm_crop_sz, bb, image.shape[1], image.shape[0],
                                           None, None, None, segm_threshold=self.params.segm_mask_thr)

        if len(cnt_area) > 0 and len(contours) != 0 and np.max(cnt_area) > 50:  # 1000:
            contour = contours[np.argmax(cnt_area)]  # use max area polygon
            polygon = contour.reshape(-1, 2)

            prbox = np.reshape(cv2.boxPoints(cv2.minAreaRect(polygon)), (4, 2))  # Rotated Rectangle

            prbox_opt = np.array([])
            if self.params.segm_optimize_polygon:
                if not self.segmentation_task:
                    mask = np.zeros(mask.shape, dtype=np.uint8)
                    cv2.drawContours(mask, [contour], -1, 1, thickness=-1)

                    # save mask to disk
                    # Note: move this below if evaluating on VOT
                    if self.params.save_mask:
                        save_mask(mask, mask_real, segm_crop_sz, bb, image.shape[1], image.shape[0],
                                  self.params.masks_save_path, self.sequence_name, self.frame_name,
                                  segm_threshold=self.params.segm_mask_thr)
                    else:
                        self.pred_mask = save_mask(mask, mask_real, segm_crop_sz, bb, image.shape[1], image.shape[0],
                                                   None, None, None, segm_threshold=self.params.segm_mask_thr)

                prbox_opt_ = fit_bbox_to_mask(mask.astype(np.int32), rotated=self.rotated_bbox)
                if prbox_opt_ is not None:
                    A1 = np.linalg.norm(np.array([prbox[0, 0], prbox[0, 1]]) - np.array([prbox[1, 0], prbox[1, 1]])) * \
                         np.linalg.norm(np.array([prbox[1, 0], prbox[1, 1]]) - np.array([prbox[2, 0], prbox[2, 1]]))
                    A_new = np.linalg.norm(np.array([prbox_opt_[0, 0], prbox_opt_[0, 1]]) - np.array(
                        [prbox_opt_[1, 0], prbox_opt_[1, 1]])) * \
                            np.linalg.norm(np.array([prbox_opt_[1, 0], prbox_opt_[1, 1]]) - np.array(
                                [prbox_opt_[2, 0], prbox_opt_[2, 1]]))
                    area_ratio = A_new / A1

                    if area_ratio > 0.1 and area_ratio < 2.5:  # 1.7
                        prbox_opt = prbox_opt_
                    else:
                        print('Bbox optimization has made too large difference.')

            displacement = np.mean(prbox, axis=0) - np.array([mask.shape[0] / 2, mask.shape[1] / 2])
            prbox = (prbox - np.mean(prbox, axis=0) + displacement) / f_ + np.array([pos[1].item(), pos[0].item()])

            self.pos[0] = np.mean(prbox[:, 1])
            self.pos[1] = np.mean(prbox[:, 0])

            if prbox_opt.size > 0:
                displacement_opt = np.mean(prbox_opt, axis=0) - np.array([mask.shape[0] / 2, mask.shape[1] / 2])
                prbox = (prbox_opt - np.mean(prbox_opt, axis=0) + displacement_opt) / f_ + np.array(
                    [pos[1].item(), pos[0].item()])

            if self.rotated_bbox:
                pred_region = [prbox[0, 0], prbox[0, 1], prbox[1, 0], prbox[1, 1], prbox[2, 0], prbox[2, 1],
                               prbox[3, 0], prbox[3, 1]]
            else:
                pred_region = [np.min(prbox[:, 0]) + 1, np.min(prbox[:, 1]) + 1,
                               np.max(prbox[:, 0]) - np.min(prbox[:, 0]) + 1,
                               np.max(prbox[:, 1]) - np.min(prbox[:, 1]) + 1]

            if not self.params.bbox_reg_for_scale:
                if len(pred_region) == 4:
                    cx_ = pred_region[0] + pred_region[2] / 2
                    cy_ = pred_region[1] + pred_region[3] / 2
                    width_ = pred_region[2]
                    height_ = pred_region[3]
                elif len(pred_region) == 8:
                    cx_ = sum(pred_region[::2]) / 4
                    cy_ = sum(pred_region[1::2]) / 4
                    width_ = max(pred_region[::2]) - min(pred_region[::2])
                    height_ = max(pred_region[1::2]) - min(pred_region[1::2])
                else:
                    print('Error: region must have 4 or 8 elements.')
                    exit(-1)

                penalty_k = 0.14
                track_lr = 0.30
                # scale penalty
                s_c = calc_change(calc_sz(width_, height_) / (calc_sz(self.bb_sz[0] * f_, self.bb_sz[1] * f_)))
                # aspect ratio penalty
                r_c = calc_change((self.bb_sz[0] / self.bb_sz[1]) / (width_ / height_))
                penalty = np.exp(-(r_c * s_c - 1) * penalty_k)
                lr = penalty * track_lr
                # smoothed width and height
                width_ = self.bb_sz[0] * (1 - lr) + width_ * lr
                height_ = self.bb_sz[1] * (1 - lr) + height_ * lr

                self.target_scale *= np.sqrt(width_ * height_) / np.sqrt(self.bb_sz[0] * self.bb_sz[1])
                self.target_scale = self.target_scale.clamp(self.min_scale_factor, self.max_scale_factor)

                self.bb_sz = np.array([width_, height_])
                self.reg_bb = [cx_ - width_ / 2, cy_ - height_ / 2, width_, height_]

                self.target_sz = self.base_target_sz * self.target_scale

            return pred_region

        return None

    def poly_to_aabbox(self, x_, y_):
        # keep the center and area of the polygon
        # change aspect ratio of the original bbox
        cx = np.mean(x_)
        cy = np.mean(y_)
        x1 = np.min(x_)
        x2 = np.max(x_)
        y1 = np.min(y_)
        y2 = np.max(y_)
        A1 = np.linalg.norm(np.array([x_[0], y_[0]]) - np.array([x_[1], y_[1]])) * \
             np.linalg.norm(np.array([x_[1], y_[1]]) - np.array([x_[2], y_[2]]))
        A2 = (x2 - x1) * (y2 - y1)
        s = np.sqrt(A1 / A2)
        w = s * (x2 - x1) + 1
        h = s * (y2 - y1) + 1
        return np.array([cx - w / 2, cy - h / 2, w, h])
